class TAFError(Exception):
    pass


class InvalidBranchError(TAFError):
    pass


class InvalidCommitError(TAFError):
    pass


class InvalidKeyError(TAFError):
    def __init__(self, metadata_role):
        super().__init__(
            f"Cannot sign {metadata_role} metadata file with inserted key."
        )


class InvalidOrMissingMetadataError(TAFError):
    pass


class InvalidRepositoryError(TAFError):
    pass


class InvalidPINError(TAFError):
    pass


class MetadataUpdateError(TAFError):
    def __init__(self, metadata_role, message):
        super().__init__(
            f"Error happened while updating {metadata_role} metadata role(s):\n\n{message}"
        )
        self.metadata_role = metadata_role
        self.message = message


class RootMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message):
        super().__init__("root", message)


class KeystoreError(TAFError):
    pass


class PINMissmatchError(Exception):
    pass


class SnapshotMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message):
        super().__init__("snapshot", message)


class TargetsMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message):
        super().__init__("targets", message)


class TimestampMetadataUpdateError(MetadataUpdateError):
    def __init__(self, message):
        super().__init__("timestamp", message)


class NoSpeculativeBranchError(TAFError):
    pass


class RepositoriesNotFoundError(TAFError):
    pass


class UpdateFailedError(TAFError):
    pass


class YubikeyError(Exception):
    pass



import datetime
from pathlib import Path

import pytest
import securesystemslib
import taf.exceptions
import tuf
from taf.tests import TEST_WITH_REAL_YK
from taf.utils import to_tuf_datetime_format


@pytest.mark.skipif(TEST_WITH_REAL_YK, reason="Testing with real Yubikey.")
def test_check_no_key_inserted_for_targets_should_raise_error(
    taf_happy_path, targets_yk
):
    targets_yk.insert()
    targets_yk.remove()
    with pytest.raises(taf.exceptions.YubikeyError):
        taf_happy_path.is_valid_metadata_yubikey("targets")


def test_check_targets_key_id_for_targets_should_return_true(
    taf_happy_path, targets_yk
):
    targets_yk.insert()
    assert taf_happy_path.is_valid_metadata_yubikey("targets", targets_yk.tuf_key)


def test_check_root_key_id_for_targets_should_return_false(taf_happy_path, root1_yk):
    root1_yk.insert()
    assert not taf_happy_path.is_valid_metadata_yubikey("targets", root1_yk.tuf_key)


def test_update_snapshot_and_timestmap(taf_happy_path, snapshot_key, timestamp_key):
    date_now = datetime.datetime.now()
    snapshot_date = date_now + datetime.timedelta(1)
    snapshot_interval = 2
    timestamp_date = date_now + datetime.timedelta(2)
    timestamp_interval = 3

    kwargs = {
        "snapshot_date": snapshot_date,
        "timestamp_date": timestamp_date,
        "snapshot_interval": snapshot_interval,
        "timestamp_interval": timestamp_interval,
    }

    taf_happy_path.update_snapshot_and_timestmap(snapshot_key, timestamp_key, **kwargs)

    targets_metadata_path = Path(taf_happy_path.metadata_path) / "targets.json"
    snapshot_metadata_path = Path(taf_happy_path.metadata_path) / "snapshot.json"
    timestamp_metadata_path = Path(taf_happy_path.metadata_path) / "timestamp.json"

    old_targets_metadata = targets_metadata_path.read_bytes()

    def check_expiration_date(metadata_path, date, interval):
        signable = securesystemslib.util.load_json_file(metadata_path)
        tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
        actual_expiration_date = signable["signed"]["expires"]

        assert actual_expiration_date == to_tuf_datetime_format(date, interval)

    check_expiration_date(str(snapshot_metadata_path), snapshot_date, snapshot_interval)
    check_expiration_date(
        str(timestamp_metadata_path), timestamp_date, timestamp_interval
    )

    # Targets data should remain the same
    assert old_targets_metadata == targets_metadata_path.read_bytes()


def test_update_snapshot_valid_key(taf_happy_path, snapshot_key):
    start_date = datetime.datetime.now()
    interval = 1
    expected_expiration_date = to_tuf_datetime_format(start_date, interval)
    taf_happy_path.update_snapshot(
        snapshot_key, start_date=start_date, interval=interval
    )
    new_snapshot_metadata = str(Path(taf_happy_path.metadata_path) / "snapshot.json")
    signable = securesystemslib.util.load_json_file(new_snapshot_metadata)
    tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
    actual_expiration_date = signable["signed"]["expires"]

    assert actual_expiration_date == expected_expiration_date


def test_update_snapshot_wrong_key(taf_happy_path, timestamp_key):
    with pytest.raises(taf.exceptions.InvalidKeyError):
        taf_happy_path.update_snapshot(timestamp_key)


def test_update_timestamp_valid_key(taf_happy_path, timestamp_key):
    start_date = datetime.datetime.now()
    interval = 1
    expected_expiration_date = to_tuf_datetime_format(start_date, interval)

    taf_happy_path.update_timestamp(
        timestamp_key, start_date=start_date, interval=interval
    )
    new_timestamp_metadata = str(Path(taf_happy_path.metadata_path) / "timestamp.json")
    signable = securesystemslib.util.load_json_file(new_timestamp_metadata)
    tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
    actual_expiration_date = signable["signed"]["expires"]

    assert actual_expiration_date == expected_expiration_date


def test_update_timestamp_wrong_key(taf_happy_path, snapshot_key):
    with pytest.raises(taf.exceptions.InvalidKeyError):
        taf_happy_path.update_timestamp(snapshot_key)


def test_update_targets_from_keystore_valid_key(taf_happy_path, targets_key):
    start_date = datetime.datetime.now()
    interval = 1
    expected_expiration_date = to_tuf_datetime_format(start_date, interval)

    taf_happy_path.update_targets_from_keystore(
        targets_key, start_date=start_date, interval=interval
    )
    new_targets_data = str(Path(taf_happy_path.metadata_path) / "targets.json")
    signable = securesystemslib.util.load_json_file(new_targets_data)
    tuf.formats.SIGNABLE_SCHEMA.check_match(signable)
    actual_expiration_date = signable["signed"]["expires"]

    assert actual_expiration_date == expected_expiration_date


def test_update_targets_from_keystore_wrong_key(taf_happy_path, snapshot_key):
    with pytest.raises(taf.exceptions.InvalidKeyError):
        taf_happy_path.update_targets_from_keystore(snapshot_key)


def test_update_targets_valid_key_valid_pin(taf_happy_path, targets_yk):
    targets_path = Path(taf_happy_path.targets_path)
    repositories_json_path = targets_path / "repositories.json"

    branch_id = "14e81cd1-0050-43aa-9e2c-e34fffa6f517"
    target_commit_sha = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
    repositories_json_old = repositories_json_path.read_text()

    targets_data = {
        "branch": {"target": branch_id},
        "dummy/target_dummy_repo": {"target": {"commit": target_commit_sha}},
        "capstone": {},
    }

    targets_yk.insert()
    taf_happy_path.update_targets(
        "123456", targets_data, datetime.datetime.now(), public_key=targets_yk.tuf_key
    )

    assert (targets_path / "branch").read_text() == branch_id
    assert target_commit_sha in (targets_path / "dummy/target_dummy_repo").read_text()
    assert (targets_path / "capstone").is_file()
    assert repositories_json_old == repositories_json_path.read_text()


def test_update_targets_valid_key_wrong_pin(taf_happy_path, targets_yk):
    with pytest.raises(taf.exceptions.TargetsMetadataUpdateError):
        targets_yk.insert()
        taf_happy_path.update_targets("123", public_key=targets_yk.tuf_key)


@pytest.mark.skipif(TEST_WITH_REAL_YK, reason="Testing with real Yubikey.")
def test_update_targets_wrong_key(taf_happy_path, root1_yk):
    with pytest.raises(taf.exceptions.InvalidKeyError):
        root1_yk.insert()
        taf_happy_path.update_targets("123456")



import pytest

from taf.exceptions import InvalidRepositoryError
from taf.git import GitRepository


def test_url_validation_valid_urls():
    urls = [
        "https://github.com/account/repo_name.git",
        "https://github.com/account/repo_name",
        "http://github.com/account/repo_name.git",
        "http://github.com/account/repo_name",
        "git@github.com:openlawlibrary/taf.git",
        "git@github.com:openlawlibrary/taf",
    ]
    repo = GitRepository("path", urls)
    for test_url, repo_url in zip(urls, repo.repo_urls):
        assert test_url == repo_url


def test_url_invalid_urls():
    urls = ["abc://something.com"]
    with pytest.raises(InvalidRepositoryError):
        GitRepository("path", urls)



import datetime
import json
import os
from binascii import hexlify
from collections import defaultdict
from functools import partial
from getpass import getpass
from pathlib import Path

import click
import securesystemslib
import securesystemslib.exceptions
import tuf.repository_tool
from taf.auth_repo import AuthenticationRepo
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import KeystoreError
from taf.git import GitRepository
from taf.keystore import (
    key_cmd_prompt,
    read_private_key_from_keystore,
    read_public_key_from_keystore,
)
from taf.log import taf_logger
from taf.repository_tool import Repository
from taf.utils import read_input_dict
from tuf.repository_tool import (
    METADATA_DIRECTORY_NAME,
    TARGETS_DIRECTORY_NAME,
    create_new_repository,
    generate_and_write_rsa_keypair,
    generate_rsa_key,
)

try:
    import taf.yubikey as yk
except ImportError:
    taf_logger.warning('"yubikey-manager" is not installed.')

# Yubikey x509 certificate expiration interval
EXPIRATION_INTERVAL = 36500
YUBIKEY_EXPIRATION_DATE = datetime.datetime.now() + datetime.timedelta(
    days=EXPIRATION_INTERVAL
)


def add_signing_key(repo_path, role, pub_key_path):
    """
    Adds a new signing key. Currently assumes that all relevant keys are stored on yubikeys.
    Allows us to add a new targets key for example
    """
    from taf.repository_tool import yubikey_signature_provider

    taf_repo = Repository(repo_path)
    pub_key_pem = Path(pub_key_path).read_text()
    taf_repo.add_metadata_key(role, pub_key_pem, DEFAULT_RSA_SIGNATURE_SCHEME)
    root_obj = taf_repo._repository.root
    threshold = root_obj.threshold
    keys_num = len(root_obj.keys)
    num_of_signatures = 0
    loaded_yubikeys = {}
    pub_key, _ = yk.yubikey_prompt(
        role, role, taf_repo, loaded_yubikeys=loaded_yubikeys
    )
    role_obj = _role_obj(role, taf_repo._repository)
    role_obj.add_external_signature_provider(
        pub_key, partial(yubikey_signature_provider, role, pub_key["keyid"])
    )

    all_loaded = False
    while not all_loaded:
        if num_of_signatures >= threshold:
            all_loaded = not (
                click.confirm(
                    "Threshold of root keys reached. Do you want to load more root keys?"
                )
            )
        if not all_loaded:
            name = f"root{num_of_signatures+1}"
            pub_key, _ = yk.yubikey_prompt(
                name, "root", taf_repo, loaded_yubikeys=loaded_yubikeys
            )
            root_obj.add_external_signature_provider(
                pub_key, partial(yubikey_signature_provider, name, pub_key["keyid"])
            )
            num_of_signatures += 1
        if num_of_signatures == keys_num:
            all_loaded = True
    taf_repo.writeall()


def _load_signing_keys(
    taf_repo,
    role,
    keystore=None,
    role_key_infos=None,
    loaded_yubikeys=None,
    scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
):
    """
    Load role's signing keys. Make sure that at least the threshold of keys was
    loaded, but allow loading more keys (so that a metadata file can be signed
    by all of the role's keys if the user wants that)
    """
    print(f"Loading signing keys of role {role}")
    threshold = taf_repo.get_role_threshold(role)
    signing_keys_num = len(taf_repo.get_role_keys(role))
    all_loaded = False
    num_of_signatures = 0
    keys = []
    # check if keystore exists and if there is a file corresponding to this role
    load_from_keystore = False
    if keystore is not None:
        keystore = Path(keystore)
        # names of keys are expected to be role or role + counter
        counter = "" if signing_keys_num == 1 else "1"
        if (keystore / f"{role}{counter}").is_file():
            load_from_keystore = True

    while not all_loaded and num_of_signatures < signing_keys_num:
        if signing_keys_num == 1:
            key_name = role
        else:
            key_name = f"{role}{num_of_signatures + 1}"
        if load_from_keystore:
            # if loading from keystore, load all keys
            key = read_private_key_from_keystore(
                keystore, key_name, role_key_infos, num_of_signatures, scheme
            )
            keys.append(key)
        else:
            if num_of_signatures >= threshold:
                all_loaded = not (
                    click.confirm(
                        f"Threshold of {role} keys reached. Do you want to load more {role} keys?"
                    )
                )
            if not all_loaded:
                is_yubikey = None
                if role_key_infos is not None and role in role_key_infos:
                    is_yubikey = role_key_infos[role].get("yubikey")
                if is_yubikey is None:
                    is_yubikey = click.confirm(f"Sign {role} using YubiKey(s)?")
                if is_yubikey:
                    yk.yubikey_prompt(
                        key_name, role, taf_repo, loaded_yubikeys=loaded_yubikeys
                    )
                else:
                    key = key_cmd_prompt(key_name, role, taf_repo, keys, scheme)
                    keys.append(key)
        num_of_signatures += 1
    return keys


def _update_target_repos(repo_path, targets_dir, target_repo_path):
    """Updates target repo's commit sha"""
    if not target_repo_path.is_dir() or target_repo_path == repo_path:
        return

    target_repo = GitRepository(str(target_repo_path))
    if target_repo.is_git_repository:
        commit = target_repo.head_commit_sha()
        target_repo_name = target_repo_path.name
        (targets_dir / target_repo_name).write_text(
            json.dumps({"commit": commit}, indent=4)
        )


def update_target_repos_from_fs(repo_path, targets_directory, namespace=None):
    """
    <Purpose>
        Create or update target files by reading the latest commits of the provided target repositories
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
    """
    repo_path = Path(repo_path).resolve()
    targets_dir = Path(targets_directory).resolve()
    if namespace is None:
        namespace = targets_dir.name
    auth_repo_targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    if namespace:
        auth_repo_targets_dir = auth_repo_targets_dir / namespace
        auth_repo_targets_dir.mkdir(parents=True, exist_ok=True)

    for target_repo_path in targets_dir.glob("*"):
        _update_target_repos(repo_path, auth_repo_targets_dir, target_repo_path)


def update_target_repos_from_repositories_json(repo_path):
    """
    <Purpose>
        Create or update target files by reading the latest commits repositories.json
    <Arguments>
        repo_path:
        Authentication repository's location
    """
    repo_path = Path(repo_path).resolve()
    targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    repositories_json = json.loads(Path(targets_dir).read_text())

    for repo in repositories_json.get("repositories"):
        target_repo_path = targets_dir / repo
        target_repo_path.mkdir(parents=True, exist_ok=True)
        _update_target_repos(repo_path, targets_dir, target_repo_path)


def build_auth_repo(
    repo_path,
    targets_directory,
    namespace,
    targets_relative_dir,
    keystore,
    roles_key_infos,
    repos_custom,
):
    # read the key infos here, no need to read the file multiple times
    roles_key_infos = read_input_dict(roles_key_infos)
    create_repository(repo_path, keystore, roles_key_infos)
    generate_repositories_json(
        repo_path, targets_directory, namespace, targets_relative_dir, repos_custom
    )
    register_target_files(
        repo_path, keystore, roles_key_infos, commit_msg="Added repositories.json"
    )
    auth_repo_targets_dir = Path(repo_path, TARGETS_DIRECTORY_NAME)
    if namespace:
        auth_repo_targets_dir = auth_repo_targets_dir / namespace
        auth_repo_targets_dir.mkdir(parents=True, exist_ok=True)
    # group commits by dates
    # first add first repos at a date, then second repost at that date
    commits_by_date = defaultdict(dict)
    target_repositories = []
    for target_repo_dir in os.listdir(targets_directory):
        target_repo = GitRepository(Path(targets_directory, target_repo_dir))
        target_repo.checkout_branch("master")
        target_repo_name = str(Path(target_repo_dir).parent)
        target_repositories.append(target_repo_name)
        commits = target_repo.list_commits(format="format:%H|%cd", date="short")
        for commit in commits[::-1]:
            sha, date = commit.split("|")
            commits_by_date[date].setdefault(target_repo_name, []).append(sha)

    for date in sorted(commits_by_date.keys()):
        repos_and_commits = commits_by_date[date]
        for target_repo_name in target_repositories:
            if target_repo_name in repos_and_commits:
                for sha in commits_by_date[date][target_repo_name]:
                    Path(auth_repo_targets_dir, target_repo_name).write_text(
                        json.dumps({"commit": sha}, indent=4)
                    )
                    register_target_files(
                        repo_path,
                        keystore,
                        roles_key_infos,
                        commit_msg=f"Updated {target_repo_name}",
                    )


def _register_yubikey(yubikeys, role_obj, role_name, key_name, scheme, certs_dir):
    from taf.repository_tool import yubikey_signature_provider

    registered = False
    while not registered:
        print(f"Registering keys for {key_name}")
        use_existing = click.confirm("Do you want to reuse already set up Yubikey?")

        if not use_existing:
            if not click.confirm(
                "WARNING - this will delete everything from the inserted key. Proceed?"
            ):
                continue

        key, serial_num = yk.yubikey_prompt(
            key_name,
            role_name,
            taf_repo=None,
            registering_new_key=True,
            creating_new_key=not use_existing,
            loaded_yubikeys=yubikeys,
            pin_confirm=True,
            pin_repeat=True,
        )

        if not use_existing:
            key = yk.setup_new_yubikey(serial_num, scheme)
        export_yk_certificate(certs_dir, key)

        registered = True
        # set Yubikey expiration date
        # add_yubikey_external_signature_provider
        role_obj.add_verification_key(key, expires=YUBIKEY_EXPIRATION_DATE)
        role_obj.add_external_signature_provider(
            key, partial(yubikey_signature_provider, key_name, key["keyid"])
        )


def _register_key(
    keystore, roles_key_info, role_obj, role_name, key_name, key_num, scheme
):
    # if keystore exists, load the keys
    if keystore is not None:
        public_key = read_public_key_from_keystore(keystore, key_name, scheme)
        private_key = read_private_key_from_keystore(
            keystore, key_name, roles_key_info, key_num, scheme
        )
    # if it does not, generate the keys and print the output
    else:
        key = generate_rsa_key(scheme=scheme)
        print(f"{role_name} key:\n\n{key['keyval']['private']}\n\n")
        public_key = private_key = key
    role_obj.add_verification_key(public_key)
    role_obj.load_signing_key(private_key)


def create_repository(
    repo_path, keystore, roles_key_infos, commit_message=None, test=False
):
    """
    <Purpose>
        Create a new authentication repository. Generate initial metadata files.
        The initial targets metadata file is empty (does not specify any targets).
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_message:
        If provided, the changes will be committed automatically using the specified message
        test:
        Indicates if the created repository is a test authentication repository
    """
    yubikeys = defaultdict(dict)
    roles_key_infos = read_input_dict(roles_key_infos)
    repo = AuthenticationRepo(repo_path)
    if Path(repo_path).is_dir():
        if repo.is_git_repository:
            print(f"Repository {repo_path} already exists")
            return

    tuf.repository_tool.METADATA_STAGED_DIRECTORY_NAME = METADATA_DIRECTORY_NAME
    repository = create_new_repository(repo_path)

    def _register_roles_keys(role_name, key_info, repository):
        num_of_keys = key_info.get("number", 1)
        threshold = key_info.get("threshold", 1)
        is_yubikey = key_info.get("yubikey", False)
        scheme = key_info.get("scheme", DEFAULT_RSA_SIGNATURE_SCHEME)

        role_obj = _role_obj(role_name, repository)
        role_obj.threshold = threshold
        for key_num in range(num_of_keys):
            key_name = _get_key_name(role_name, key_num, num_of_keys)
            if is_yubikey:
                _register_yubikey(
                    yubikeys, role_obj, role_name, key_name, scheme, repo.certs_dir
                )
            else:
                _register_key(
                    keystore, key_info, role_obj, role_name, key_name, key_num, scheme
                )

    try:
        # load keys not stored on YubiKeys first, to avoid entering pins
        # if there is somethig wrong with keystore files
        for role_name, key_info in roles_key_infos.items():
            if not key_info.get("yubikey", False):
                _register_roles_keys(role_name, key_info, repository)

        for role_name, key_info in roles_key_infos.items():
            if key_info.get("yubikey", False):
                _register_roles_keys(role_name, key_info, repository)

    except KeystoreError as e:
        print(f"Creation of repository failed: {e}")
        return

    # if the repository is a test repository, add a target file called test-auth-repo
    if test:
        target_paths = Path(repo_path) / "targets"
        test_auth_file = target_paths / "test-auth-repo"
        test_auth_file.touch()
        targets_obj = _role_obj("targets", repository)
        targets_obj.add_target(str(test_auth_file))

    repository.writeall()
    if commit_message is not None and len(commit_message):
        auth_repo = GitRepository(repo_path)
        auth_repo.init_repo()
        auth_repo.commit(commit_message)


def export_yk_public_pem(path=None):
    pub_key_pem = yk.export_piv_pub_key().decode("utf-8")
    if path is None:
        print(pub_key_pem)
    else:
        if not path.endswith(".pub"):
            path = f"{path}.pub"
        path = Path(path)
        parent = path.parent
        parent.mkdir(parents=True, exist_ok=True)
        path.write_text(pub_key_pem)


def export_yk_certificate(certs_dir, key):
    if certs_dir is None:
        certs_dir = Path.home()
    else:
        certs_dir = Path(certs_dir)
    certs_dir.mkdir(parents=True, exist_ok=True)
    cert_path = certs_dir / f"{key['keyid']}.cert"
    print(f"Exporting certificate to {cert_path}")
    with open(cert_path, "wb") as f:
        f.write(yk.export_piv_x509())


def generate_keys(keystore, roles_key_infos):
    """
    <Purpose>
        Generate public and private keys and writes them to disk. Names of keys correspond to names
        of the TUF roles. If more than one key should be generated per role, a counter is appended
        to the role's name. E.g. root1, root2, root3 etc.
    <Arguments>
        keystore:
        Location where the generated files should be saved
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        This includes:
            - passwords of the keystore files
            - number of keys per role (optional, defaults to one if not provided)
            - key length (optional, defaults to TUF's default value, which is 3072)
        Names of the keys are set to names of the roles plus a counter, if more than one key
        should be generated.
    """
    roles_key_infos = read_input_dict(roles_key_infos)
    for role_name, key_info in roles_key_infos.items():
        num_of_keys = key_info.get("number", 1)
        bits = key_info.get("length", 3072)
        passwords = key_info.get("passwords", [""] * num_of_keys)
        is_yubikey = key_info.get("yubikey", False)
        for key_num in range(num_of_keys):
            if not is_yubikey:
                key_name = _get_key_name(role_name, key_num, num_of_keys)
                password = passwords[key_num]
                generate_and_write_rsa_keypair(
                    str(Path(keystore, key_name)), bits=bits, password=password
                )


def generate_repositories_json(
    repo_path,
    targets_directory,
    namespace=None,
    targets_relative_dir=None,
    custom_data=None,
):
    """
    <Purpose>
        Generatesinitial repositories.json
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
        targets_relative_dir:
        Directory relative to which urls of the target repositories are set, if they do not have remote set
    """

    custom_data = read_input_dict(custom_data)
    repositories = {}

    repo_path = Path(repo_path).resolve()
    auth_repo_targets_dir = repo_path / TARGETS_DIRECTORY_NAME
    targets_directory = Path(targets_directory).resolve()
    if targets_relative_dir is not None:
        targets_relative_dir = Path(targets_relative_dir).resolve()
    if namespace is None:
        namespace = targets_directory.name
    for target_repo_dir in targets_directory.glob("*"):
        if not target_repo_dir.is_dir() or target_repo_dir == repo_path:
            continue
        target_repo = GitRepository(target_repo_dir.resolve())
        if not target_repo.is_git_repository:
            continue
        target_repo_name = target_repo_dir.name
        target_repo_namespaced_name = (
            target_repo_name if not namespace else f"{namespace}/{target_repo_name}"
        )
        # determine url to specify in initial repositories.json
        # if the repository has a remote set, use that url
        # otherwise, set url to the repository's absolute or relative path (relative
        # to targets_relative_dir if it is specified)
        url = target_repo.get_remote_url()
        if url is None:
            if targets_relative_dir is not None:
                url = Path(target_repo.repo_path).relative_to(targets_relative_dir)
            else:
                url = Path(target_repo.repo_path).resolve()
            # convert to posix path
            url = str(url.as_posix())
        repositories[target_repo_namespaced_name] = {"urls": [url]}
        if target_repo_namespaced_name in custom_data:
            repositories[target_repo_namespaced_name]["custom"] = custom_data[
                target_repo_namespaced_name
            ]

    (auth_repo_targets_dir / "repositories.json").write_text(
        json.dumps({"repositories": repositories}, indent=4)
    )


def _get_key_name(role_name, key_num, num_of_keys):
    if num_of_keys == 1:
        return role_name
    else:
        return role_name + str(key_num + 1)


def init_repo(
    repo_path,
    targets_directory,
    namespace,
    targets_relative_dir,
    keystore,
    roles_key_infos,
    repos_custom=None,
    commit=None,
    test=False,
):
    """
    <Purpose>
        Generate initial repository:
        1. Crete tuf authentication repository
        2. Commit initial metadata files if commit == True
        3. Add target repositories
        4. Generate repositories.json
        5. Update tuf metadata
        6. Commit the changes if commit == True
    <Arguments>
        repo_path:
        Authentication repository's location
        targets_directory:
        Directory which contains target repositories
        namespace:
        Namespace used to form the full name of the target repositories. E.g. some_namespace/law-xml
        targets_relative_dir:
        Directory relative to which urls of the target repositories are set, if they do not have remote set
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_message:
        If provided, the changes will be committed automatically using the specified message
        test:
        Indicates if the created repository is a test authentication repository
    """
    # read the key infos here, no need to read the file multiple times
    roles_key_infos = read_input_dict(roles_key_infos)
    commit_msg = "Initial commit" if commit else None
    create_repository(repo_path, keystore, roles_key_infos, commit_msg, test)
    update_target_repos_from_fs(repo_path, targets_directory, namespace)
    generate_repositories_json(
        repo_path, targets_directory, namespace, targets_relative_dir, repos_custom
    )
    register_target_files(repo_path, keystore, roles_key_infos, commit_msg=commit)


def register_target_file(repo_path, file_path, keystore, roles_key_infos, scheme):
    roles_key_infos = read_input_dict(roles_key_infos)
    taf_repo = Repository(repo_path)
    taf_repo.add_existing_target(file_path)

    _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme)


def register_target_files(
    repo_path,
    keystore,
    roles_key_infos,
    commit_msg=None,
    scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
):
    """
    <Purpose>
        Register all files found in the target directory as targets - updates the targets
        metadata file, snapshot and timestamp. Sign targets
        with yubikey if keystore is not provided
    <Arguments>
        repo_path:
        Authentication repository's path
        keystore:
        Location of the keystore files
        roles_key_infos:
        A dictionary whose keys are role names, while values contain information about the keys.
        commit_msg:
        Commit message. If specified, the changes made to the authentication are committed.
        scheme:
        A signature scheme used for signing.
    """
    roles_key_infos = read_input_dict(roles_key_infos)
    repo_path = Path(repo_path).resolve()
    targets_path = repo_path / TARGETS_DIRECTORY_NAME
    taf_repo = Repository(str(repo_path))
    for root, _, filenames in os.walk(str(targets_path)):
        for filename in filenames:
            taf_repo.add_existing_target(str(Path(root) / filename))
    _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme)
    if commit_msg is not None:
        auth_repo = GitRepository(repo_path)
        auth_repo.commit(commit_msg)


def _role_obj(role, repository):
    if role == "targets":
        return repository.targets
    elif role == "snapshot":
        return repository.snapshot
    elif role == "timestamp":
        return repository.timestamp
    elif role == "root":
        return repository.root


def signature_provider(key_id, cert_cn, key, data):  # pylint: disable=W0613
    def _check_key_id(expected_key_id):
        try:
            inserted_key = yk.get_piv_public_key_tuf()
            return expected_key_id == inserted_key["keyid"]
        except Exception:
            return False

    while not _check_key_id(key_id):
        pass

    data = securesystemslib.formats.encode_canonical(data).encode("utf-8")
    key_pin = getpass(f"Please insert {cert_cn} YubiKey, input PIN and press ENTER.\n")
    signature = yk.sign_piv_rsa_pkcs1v15(data, key_pin)

    return {"keyid": key_id, "sig": hexlify(signature).decode()}


def setup_signing_yubikey(certs_dir=None, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
    _, serial_num = yk.yubikey_prompt(
        "new Yubikey",
        creating_new_key=True,
        pin_confirm=True,
        pin_repeat=True,
        prompt_message="Please insert the new Yubikey and press ENTER",
    )
    key = yk.setup_new_yubikey(serial_num)
    export_yk_certificate(certs_dir, key)


def update_metadata_expiration_date(
    repo_path,
    role,
    key,
    start_date=datetime.datetime.now(),
    interval=None,
    commit_msg=None,
):
    taf_repo = Repository(repo_path)
    update_methods = {
        "timestamp": taf_repo.update_timestamp,
        "snapshot": taf_repo.update_snapshot,
        "targets": taf_repo.update_targets_from_keystore,
    }
    update_methods[role](key, start_date, interval)

    if commit_msg is not None:
        auth_repo = GitRepository(repo_path)
        auth_repo.commit(commit_msg)


def _write_targets_metadata(taf_repo, keystore, roles_key_infos, scheme):

    loaded_yubikeys = {}
    targets_keys = _load_signing_keys(
        taf_repo, "targets", keystore, roles_key_infos, loaded_yubikeys
    )
    if len(targets_keys):
        taf_repo.update_targets_from_keystore(targets_keys[0], write=False)
    else:
        for serial_num in loaded_yubikeys:
            if "targets" in loaded_yubikeys[serial_num]:
                pin = yk.get_key_pin(serial_num)
                taf_repo.update_targets(pin, write=False)
                break
    snapshot_keys = _load_signing_keys(taf_repo, "snapshot", keystore, roles_key_infos)
    timestamp_keys = _load_signing_keys(
        taf_repo, "timestamp", keystore, roles_key_infos
    )
    snapshot_key = snapshot_keys[0]
    timestamp_key = timestamp_keys[0]
    taf_repo.update_snapshot_and_timestmap(snapshot_key, timestamp_key, write=False)
    taf_repo.writeall()


# TODO Implement update of repositories.json (updating urls, custom data, adding new repository, removing
# repository etc.)
# TODO create tests for this



import datetime
import json
import os
import stat
import subprocess
from getpass import getpass
from pathlib import Path

import click
import taf.settings
from taf.exceptions import PINMissmatchError
from taf.log import taf_logger


def _iso_parse(date):
    return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S.%f")


class IsoDateParamType(click.ParamType):
    name = "iso_date"

    def convert(self, value, param, ctx):
        if value is None:
            return datetime.datetime.now()

        if isinstance(value, datetime.datetime):
            return value
        try:
            return _iso_parse(value)
        except ValueError as ex:
            self.fail(str(ex), param, ctx)


ISO_DATE_PARAM_TYPE = IsoDateParamType()


def extract_x509(cert_pem):
    from cryptography import x509
    from cryptography.hazmat.backends import default_backend

    cert = x509.load_pem_x509_certificate(cert_pem, default_backend())

    def _get_attr(oid):
        attrs = cert.subject.get_attributes_for_oid(oid)
        return attrs[0].value if len(attrs) > 0 else ""

    return {
        "name": _get_attr(x509.OID_COMMON_NAME),
        "organization": _get_attr(x509.OID_ORGANIZATION_NAME),
        "country": _get_attr(x509.OID_COUNTRY_NAME),
        "state": _get_attr(x509.OID_STATE_OR_PROVINCE_NAME),
        "locality": _get_attr(x509.OID_LOCALITY_NAME),
        "valid_from": cert.not_valid_before.strftime("%Y-%m-%d"),
        "valid_to": cert.not_valid_after.strftime("%Y-%m-%d"),
    }


def get_cert_names_from_keyids(certs_dir, keyids):
    cert_names = []
    for keyid in keyids:
        try:
            name = extract_x509((Path(certs_dir) / keyid + ".pem").read_bytes())["name"]
            if not name:
                print("Cannot extract common name from x509, using key id instead.")
                cert_names.append(keyid)
            else:
                cert_names.append(name)
        except FileNotFoundError:
            print(f"Certificate does not exist ({keyid}).")
    return cert_names


def get_pin_for(name, confirm=True, repeat=True):
    pin = getpass(f"Enter PIN for {name}: ")
    if confirm:
        if pin != getpass(f"Confirm PIN for {name}: "):
            err_msg = "PINs don't match!"
            if repeat:
                print(err_msg)
                get_pin_for(name, confirm, repeat)
            else:
                raise PINMissmatchError(err_msg)
    return pin


def read_input_dict(value):
    if value is None:
        return {}
    if type(value) is str:
        if Path(value).is_file():
            with open(value) as f:
                value = json.loads(f.read())
        else:
            value = json.loads(value)
    return value


def run(*command, **kwargs):
    """Run a command and return its output. Call with `debug=True` to print to
    stdout."""
    if len(command) == 1 and isinstance(command[0], str):
        command = command[0].split()
    if taf.settings.LOG_COMMAND_OUTPUT:
        taf_logger.debug("About to run command {}", " ".join(command))

    def _format_word(word, **env):
        """To support word such as @{u} needed for git commands."""
        try:
            return word.format(env)
        except KeyError:
            return word

    command = [_format_word(word, **os.environ) for word in command]
    try:
        options = dict(
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            universal_newlines=True,
        )
        options.update(kwargs)
        completed = subprocess.run(command, **options)
    except subprocess.CalledProcessError as err:
        if err.stdout:
            taf_logger.debug(err.stdout)
        if err.stderr:
            taf_logger.debug(err.stderr)
        taf_logger.info(
            "Command {} returned non-zero exit status {}",
            " ".join(command),
            err.returncode,
        )
        raise err
    if completed.stdout:
        if taf.settings.LOG_COMMAND_OUTPUT:
            taf_logger.debug(completed.stdout)
    return completed.stdout.rstrip() if completed.returncode == 0 else None


def normalize_file_line_endings(file_path):
    with open(file_path, "rb") as open_file:
        content = open_file.read()
    replaced_content = normalize_line_endings(content)
    if replaced_content != content:
        with open(file_path, "wb") as open_file:
            open_file.write(replaced_content)


def normalize_line_endings(file_content):
    WINDOWS_LINE_ENDING = b"\r\n"
    UNIX_LINE_ENDING = b"\n"
    replaced_content = file_content.replace(
        WINDOWS_LINE_ENDING, UNIX_LINE_ENDING
    ).rstrip(UNIX_LINE_ENDING)
    return replaced_content


def on_rm_error(_func, path, _exc_info):
    """Used by when calling rmtree to ensure that readonly files and folders
    are deleted.
    """
    os.chmod(path, stat.S_IWRITE)
    os.unlink(path)


def to_tuf_datetime_format(start_date, interval):
    """Used to convert datetime to format used while writing metadata:
    e.g. "2020-05-29T21:59:34Z",
    """
    datetime_object = start_date + datetime.timedelta(interval)
    datetime_object = datetime_object.replace(microsecond=0)
    return datetime_object.isoformat() + "Z"



import os
import shutil
import tempfile
from pathlib import Path
from subprocess import CalledProcessError

import securesystemslib
import tuf.client.handlers as handlers

from taf.log import taf_logger
import taf.settings as settings
from taf.auth_repo import AuthenticationRepo, NamedAuthenticationRepo
from taf.exceptions import UpdateFailedError
from taf.git import GitRepository
from taf.utils import on_rm_error


class GitUpdater(handlers.MetadataUpdater):
    """
    This class implements parts of the update process specific to keeping
    metadata files and targets in a git repository. The vast majority of the
    update process is handled by TUF's updater. This class does not modify
    and of TUF's validation functionalities - it only handles loading of
    files.
    One substantial difference between our system and how TUF is designed lies in the
    fact that we want to validate every commit and not just check if the most recent
    one contains valid metadata. We want to check if the metadata was valid at the time
    when it was committed.

    Since we do not want to change the clients repository until
    we know it is safe to use git pull, the update works as follows:
    - The repository containing the metadata files is cloned to the temp folder.
    - The most recent commit in the client's local repository is determined. Based
    on that all new commits (newer than the client's most recent one in the
    cloned repository) are found.
    - A commit is considered to be a TUF mirror. We keep track of the current commit.
    - This class is designed in such a way that for each subsequent call of the
    updater's refresh method the next commit is used as a mirror. This is better than
    instantiating this class multiple times as that seems to require less modifications
    of TUF's code.
    - The updater's method '_get_metadata_file' call 'get_mirrors'. It then iterates
    through these mirrors and tries to update a metadata file by downloading them
    from each mirror, until a valid metadata is downloaded. If none of the mirrors
    contains valid metadata, an exception is raised. So, what we want to do is to
    return the current commit, and just the current commit. This means that that an
    exception will be raised if the version of the metadata file at that commit
    is not valid.
    - The same logic is used to handle targets.


    Attributes:
        - repository_directory: the client's local repository's location
        - current_path: path of the 'current' directory needed by the updater
        - previous_path: path of the 'previous' directory needed by the updater
        - validation_auth_repo: a fresh clone of the metadata repository. It is
        a bare git repository. An instance of the `BareGitRepo` class.
        - users_auth_repo: an instance of the `GitRepo` class. The user's current
        git repository.
        - commits: a list of commits, starting with the most recent commit in the
        user's repository. The following commits are those committed after the
        the top one if the client's repo.
        - commits_indexes: a dictionary which stores index of the current commit
        per metadata file. The reason for separating the metadata files is that
        not all files are updated at the same time.
    """

    @property
    def current_commit(self):
        return self.commits[self.current_commit_index]

    @property
    def previous_commit(self):
        return self.commits[self.current_commit_index - 1]

    def __init__(self, mirrors, repository_directory, repository_name):
        """
        Args:
        mirrors: is a dictionary which contains information about each mirror:
                    mirrors = {'mirror1': {'url_prefix': 'http://localhost:8001',
                            'metadata_path': 'metadata',
                            'targets_path': 'targets',
                            'confined_target_dirs': ['']}}
        This dictionary is provided by the user of the updater and used to
        create an instance of the tuf updater.
        We use url_prefix to specify url of the git repository which we want to clone.
        repository_directory: the client's local repository's location
        """
        super(GitUpdater, self).__init__(mirrors, repository_directory, repository_name)

        auth_url = mirrors["mirror1"]["url_prefix"]
        self.metadata_path = mirrors["mirror1"]["metadata_path"]
        self.targets_path = mirrors["mirror1"]["targets_path"]
        if settings.validate_repo_name:
            self.users_auth_repo = NamedAuthenticationRepo(
                repository_directory,
                repository_name,
                self.metadata_path,
                self.targets_path,
                repo_urls=[auth_url],
            )
        else:
            users_repo_path = Path(repository_directory, repository_name)
            self.users_auth_repo = AuthenticationRepo(
                str(users_repo_path),
                self.metadata_path,
                self.targets_path,
                repo_urls=[auth_url],
            )

        self._clone_validation_repo(auth_url)
        repository_directory = Path(self.users_auth_repo.repo_path)
        if repository_directory.exists():
            if not self.users_auth_repo.is_git_repository_root:
                if repository_directory.glob("*"):
                    raise UpdateFailedError(
                        f"{repository_directory} is not a git repository and is not empty"
                    )

        # validation_auth_repo is a freshly cloned bare repository.
        # It is cloned to a temporary directory that should be removed
        # once the update is completed

        self._init_commits()
        # users_auth_repo is the authentication repository
        # located on the users machine which needs to be updated
        self.repository_directory = str(repository_directory)

        self._init_metadata()

    def _init_commits(self):
        """
        Given a client's local repository which needs to be updated, creates a list
        of commits of the authentication repository newer than the most recent
        commit of the client's repository. These commits need to be validated.
        If the client's repository does not exist, all commits should be validated.
        We have to presume that the initial metadata is correct though (or at least
        the initial root.json).
        """
        # TODO check if users authentication repository is clean

        # load the last validated commit from he conf file
        last_validated_commit = self.users_auth_repo.last_validated_commit

        try:
            commits_since = self.validation_auth_repo.all_commits_since_commit(
                last_validated_commit
            )
        except CalledProcessError as e:
            if "Invalid revision range" in e.output:
                taf_logger.error(
                    "Commit {} is not contained by the remote repository {}.",
                    last_validated_commit,
                    self.validation_auth_repo.repo_name,
                )
                raise UpdateFailedError(
                    f"Commit {last_validated_commit} is no longer contained by repository"
                    f" {self.validation_auth_repo.repo_name}. This could "
                    "either mean that there was an unauthorized push tot the remote "
                    "repository, or that last_validated_commit file was modified."
                )
            else:
                raise e

        # Check if the user's head commit matches the saved one
        # That should always be the case
        # If it is not, it means that someone, accidentally or maliciously made manual changes

        if not self.users_auth_repo.is_git_repository_root:
            users_head_sha = None
        else:
            self.users_auth_repo.checkout_branch("master")
            if last_validated_commit is not None:
                users_head_sha = self.users_auth_repo.head_commit_sha()
            else:
                # if the user's repository exists, but there is no last_validated_commit
                # start the update from the beginning
                users_head_sha = None

        if last_validated_commit != users_head_sha:
            # TODO add a flag --force/f which, if provided, should force an automatic revert
            # of the users authentication repository to the last validated commit
            # This could be done if a user accidentally committed something to the auth repo
            # or manually pulled the changes
            # If the user deleted the repository or executed reset --hard, we could handle
            # that by starting validation from the last validated commit, as opposed to the
            # user's head sha.
            # For now, we will raise an error
            msg = (
                f"Saved last validated commit {last_validated_commit} does not match the head"
                f" commit of the authentication repository {users_head_sha}"
            )
            taf_logger.error(msg)
            raise UpdateFailedError(msg)

        # insert the current one at the beginning of the list
        if users_head_sha is not None:
            commits_since.insert(0, users_head_sha)

        self.commits = commits_since
        self.users_head_sha = users_head_sha
        self.current_commit_index = 0

    def _init_metadata(self):
        """
        TUF updater expects the existence of two directories in the client's
        metadata directory - current and previous. These directories store
        the current and previous metadata files (before and after the update).
        They must exists and contain at least root.json. Otherwise, update will
        fail. We actually want to validate the remote authentication repository,
        but will create the directories where TUF expects them to be in order
        to avoid modifying the updater.
        """
        # create current and previous directories and copy the metadata files
        # needed by the updater
        # TUF's updater expects these directories to be in the client's repository
        # read metadata of the cloned validation repo at the initial commit

        metadata_path = Path(self.repository_directory, "metadata")
        metadata_path.mkdir(parents=True, exist_ok=True)
        self.current_path = Path(metadata_path, "current")
        self.current_path.mkdir()
        self.previous_path = Path(metadata_path, "previous")
        self.previous_path.mkdir()

        metadata_files = self.validation_auth_repo.list_files_at_revision(
            self.current_commit, "metadata"
        )
        for filename in metadata_files:
            metadata = self.validation_auth_repo.get_file(
                self.current_commit, "metadata/" + filename
            )
            current_filename = self.current_path / filename
            previous_filename = self.previous_path / filename
            current_filename.write_text(metadata)
            shutil.copyfile(str(current_filename), str(previous_filename))

    def _clone_validation_repo(self, url):
        """
        Clones the authentication repository based on the url specified using the
        mirrors parameter. The repository is cloned as a bare repository
        to a the temp directory and will be deleted one the update is done.
        """
        temp_dir = tempfile.mkdtemp()
        repo_path = Path(temp_dir, self.users_auth_repo.repo_name).absolute()
        self.validation_auth_repo = GitRepository(repo_path, [url])
        self.validation_auth_repo.clone(bare=True)
        self.validation_auth_repo.fetch(fetch_all=True)

    def cleanup(self):
        """
        Removes the bare authentication repository and current and previous
        directories. This should be called after the update is finished,
        either successfully or unsuccessfully.
        """
        shutil.rmtree(self.current_path)
        shutil.rmtree(self.previous_path)
        temp_dir = Path(self.validation_auth_repo.repo_path, os.pardir).parent
        shutil.rmtree(str(temp_dir), onerror=on_rm_error)

    def earliest_valid_expiration_time(self):
        # metadata at a certain revision should not expire before the
        # time it was committed. It can be expected that the metadata files
        # at older commits will be expired and that should not be considered
        # to be an error
        return int(self.validation_auth_repo.get_commits_date(self.current_commit))

    def ensure_not_changed(self, metadata_filename):
        """
        Make sure that the metadata file remained the same, as the reference metadata suggests.
        """
        current_file = self.get_metadata_file(
            self.current_commit, file_name=metadata_filename
        )
        previous_file = self.get_metadata_file(
            self.previous_commit, file_name=metadata_filename
        )
        if current_file.read() != previous_file.read():
            raise UpdateFailedError(
                f"Metadata file {metadata_filename} should be the same at revisions"
                f" {self.previous_commit} and {self.current_commit}, but is not."
            )

    def get_current_targets(self):
        return self.validation_auth_repo.list_files_at_revision(
            self.current_commit, "targets"
        )

    def get_mirrors(self, _file_type, _file_path):
        # pylint: disable=unused-argument
        # return a list containing just the current commit
        return [self.current_commit]

    def get_metadata_file(self, file_mirror, file_name, _upperbound_filelength=None):
        return self._get_file(file_mirror, "metadata/" + file_name)

    def get_target_file(self, file_mirror, _file_length, _download_safely, file_path):
        return self._get_file(file_mirror, "targets/" + file_path)

    def _get_file(self, commit, filepath):
        f = self.validation_auth_repo.get_file(commit, filepath)
        temp_file_object = securesystemslib.util.TempFile()
        temp_file_object.write(f.encode())
        return temp_file_object

    def get_file_digest(self, filepath, algorithm):
        filepath = Path(filepath).relative_to(self.validation_auth_repo.get_file)
        file_obj = self._get_file(self.current_commit, filepath)
        return securesystemslib.hash.digest_fileobject(file_obj, algorithm=algorithm)

    def on_successful_update(self, filename, mirror):
        # after the is successfully completed, set the
        # next commit as current for the given file
        taf_logger.debug("{} updated from commit {}", filename, mirror)

    def on_unsuccessful_update(self, filename):
        taf_logger.error("Failed to update {}", filename)

    def update_done(self):
        # the only metadata file that is always updated
        # regardless of if it changed or not is timestamp
        # so we can check if timestamp was updated a certain
        # number of times
        self.current_commit_index += 1
        return self.current_commit_index == len(self.commits)



from getpass import getpass
from pathlib import Path

import click
import securesystemslib
from securesystemslib.interface import (
    import_rsa_privatekey_from_file,
    import_rsa_publickey_from_file,
)
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import KeystoreError
from tuf.repository_tool import import_rsakey_from_pem


def _form_private_pem(pem):
    return f"-----BEGIN RSA PRIVATE KEY-----\n{pem}\n-----END RSA PRIVATE KEY-----"


def key_cmd_prompt(
    key_name, role, taf_repo, loaded_keys=None, scheme=DEFAULT_RSA_SIGNATURE_SCHEME
):
    def _enter_and_check_key(key_name, role, loaded_keys, scheme):
        pem = getpass(f"Enter {key_name} private key without its header and footer\n")
        pem = _form_private_pem(pem)
        try:
            key = import_rsakey_from_pem(pem, scheme)
        except Exception:
            print("Invalid key")
            return None
        public_key = import_rsakey_from_pem(key["keyval"]["public"], scheme)
        if not taf_repo.is_valid_metadata_yubikey(role, public_key):
            print(f"The entered key is not a valid {role} key")
            return None
        if loaded_keys is not None and key in loaded_keys:
            print("Key already entered")
            return None
        return key

    while True:
        key = _enter_and_check_key(key_name, role, loaded_keys, scheme)
        if key is not None:
            return key


def load_tuf_private_key(key_str, key_name, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
    if not key_str:
        key_str = getpass(
            f"Enter {key_name} private key without its header and footer\n"
        )
    key_pem = _form_private_pem(key_str)

    return import_rsakey_from_pem(key_pem, scheme)


def read_private_key_from_keystore(
    keystore,
    key_name,
    roles_key_info=None,
    key_num=None,
    scheme=DEFAULT_RSA_SIGNATURE_SCHEME,
):
    key_path = Path(keystore, key_name)
    if not key_path.is_file():
        raise KeystoreError(f"{str(key_path)} does not exist")

    password = None
    if roles_key_info is not None:
        passwords = roles_key_info.get("passwords")
        if passwords is not None and len(passwords):
            password = passwords[key_num]

    def _read_key(path, password):
        if password is None:
            password = getpass(
                f"Enter {key_name} keystore file password and press ENTER"
            )
        if not password:
            password = None
        try:
            return import_rsa_privatekey_from_file(
                str(Path(keystore, key_name)), password, scheme=scheme
            )
        except (
            securesystemslib.exceptions.FormatError,
            securesystemslib.exceptions.Error,
        ) as e:
            if "password" in str(e).lower():
                return None
            raise KeystoreError(e)
        except Exception:
            return None

    while True:
        key = _read_key(key_path, password)
        if key is not None:
            return key
        if not click.confirm("Could not open keystore file. Try again?"):
            raise KeystoreError(f"Could not open keystore file {key_path}")


def read_public_key_from_keystore(
    keystore, key_name, scheme=DEFAULT_RSA_SIGNATURE_SCHEME
):
    pub_key_path = Path(keystore, f"{key_name}.pub")
    if not pub_key_path.is_file():
        raise KeystoreError(f"{str(pub_key_path)} does not exist")
    try:
        return import_rsa_publickey_from_file(str(pub_key_path), scheme)
    except (
        securesystemslib.exceptions.FormatError,
        securesystemslib.exceptions.Error,
    ) as e:
        raise KeystoreError(e)



import json
import os
import re
import shutil
import subprocess
from collections import OrderedDict
from pathlib import Path
from taf.log import taf_logger
import taf.settings as settings
from taf.exceptions import InvalidRepositoryError
from taf.utils import run

EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"


class GitRepository(object):
    def __init__(
        self, repo_path, repo_urls=None, additional_info=None, default_branch="master"
    ):
        """
    Args:
      repo_path: repository's path
      repo_urls: repository's urls (optional)
      additional_info: a dictionary containing other data (optional)
      default_branch: repository's default branch
    """
        self.repo_path = str(repo_path)
        self.default_branch = default_branch
        if repo_urls is not None:
            if settings.update_from_filesystem is False:
                for url in repo_urls:
                    _validate_url(url)
            else:
                repo_urls = [
                    os.path.normpath(os.path.join(self.repo_path, url))
                    if not os.path.isabs(url)
                    else url
                    for url in repo_urls
                ]
        self.repo_urls = repo_urls
        self.additional_info = additional_info
        self.repo_name = os.path.basename(self.repo_path)

    _remotes = None

    @property
    def remotes(self):
        if self._remotes is None:
            self._remotes = self._git("remote").split("\n")
        return self._remotes

    @property
    def is_git_repository_root(self):
        """Check if path is git repository."""
        git_path = Path(self.repo_path) / ".git"
        return self.is_git_repository and (git_path.is_dir() or git_path.is_file())

    @property
    def is_git_repository(self):
        try:
            self._git("rev-parse --git-dir")
            return True
        except subprocess.CalledProcessError:
            return False

    @property
    def initial_commit(self):
        return (
            self._git("rev-list --max-parents=0 HEAD").strip()
            if self.is_git_repository
            else None
        )

    def is_remote_branch(self, branch_name):
        for remote in self.remotes:
            if branch_name.startswith(remote + "/"):
                return True
        return False

    def _git(self, cmd, *args, **kwargs):
        """Call git commands in subprocess
    e.g.:
      self._git('checkout {}', branch_name)
    """
        log_error = kwargs.pop("log_error", False)
        log_error_msg = kwargs.pop("log_error_msg", "")
        reraise_error = kwargs.pop("reraise_error", False)
        log_success_msg = kwargs.pop("log_success_msg", "")

        if len(args):
            cmd = cmd.format(*args)
        command = f"git -C {self.repo_path} {cmd}"
        if log_error or log_error_msg:
            try:
                result = run(command)
                if log_success_msg:
                    taf_logger.debug("Repo {}:" + log_success_msg, self.repo_name)
            except subprocess.CalledProcessError as e:
                if log_error_msg:
                    taf_logger.error(log_error_msg)
                else:
                    taf_logger.error(
                        "Repo {}: error occurred while executing {}:\n{}",
                        self.repo_name,
                        command,
                        e.output,
                    )
                if reraise_error:
                    raise
        else:
            result = run(command)
            if log_success_msg:
                taf_logger.debug("Repo {}: " + log_success_msg, self.repo_name)
        return result

    def all_commits_on_branch(self, branch=None, reverse=True):
        """Returns a list of all commits on the specified branch. If branch is None,
        all commits on the currently checked out branch will be returned
        """
        if branch is None:
            branch = ""
        commits = self._git("log {} --format=format:%H", branch).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            if reverse:
                commits.reverse()

        taf_logger.debug(
            "Repo {}: found the following commits: {}",
            self.repo_name,
            ", ".join(commits),
        )
        return commits

    def all_commits_since_commit(self, since_commit, reverse=True):
        """Returns a list of all commits since the specified commit on the
        currently checked out branch
        """
        if since_commit is None:
            return self.all_commits_on_branch(reverse=reverse)
        commits = self._git("rev-list {}..HEAD", since_commit).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            if reverse:
                commits.reverse()

        taf_logger.debug(
            "Repo {}: found the following commits after commit {}: {}",
            self.repo_name,
            since_commit,
            ", ".join(commits),
        )
        return commits

    def all_fetched_commits(self, branch="master"):
        commits = self._git("rev-list ..origin/{}", branch).strip()
        if not commits:
            commits = []
        else:
            commits = commits.split("\n")
            commits.reverse()
        taf_logger.debug(
            "Repo {}: fetched the following commits {}",
            self.repo_name,
            ", ".join(commits),
        )
        return commits

    def branches(self):
        """Returns all branches."""
        return [
            branch.strip('"').strip("'").strip()
            for branch in self._git("branch --format='%(refname:short)'").split("\n")
        ]

    def branches_containing_commit(self, commit, strip_remote=False, sort_key=None):
        """Finds all branches that contain the given commit"""
        local_branches = self._git(f"branch --contains {commit}").split("\n")
        if local_branches:
            local_branches = [
                branch.replace("*", "").strip() for branch in local_branches
            ]
        remote_branches = self._git(f"branch -r --contains {commit}").split("\n")
        filtered_remote_branches = []
        if remote_branches:
            for branch in remote_branches:
                branch = branch.strip()
                local_name = self.branch_local_name(branch)
                if (
                    local_name
                    and "HEAD" not in local_name
                    and local_name not in local_branches
                ):
                    if strip_remote:
                        branch = self.branch_local_name(branch)
                    filtered_remote_branches.append(branch)
        branches = {branch: False for branch in local_branches}
        branches.update({branch: True for branch in filtered_remote_branches})
        return OrderedDict(sorted(branches.items(), key=sort_key, reverse=True))

    def branch_exists(self, branch_name, include_remotes=True):
        """
        Checks if a branch with the given name exists.
        If include_remotes is set to True, this checks if
        a remote branch exists.
        """
        branch = self._git(f"branch --list {branch_name}")
        # this git command should return the branch's name if it exists
        # empty string otherwise
        if branch:
            return True
        if include_remotes:
            for remote in self.remotes:
                remote_branch = self._git(f"branch -r --list {remote}/{branch_name}")
                # this command should return the branch's name if the remote branch
                # exists
                # it will also return some warnings if there are problems with some refs
                if branch_name in remote_branch:
                    return True
        return False

    def branch_off_commit(self, branch_name, commit):
        """Create a new branch by branching off of the specified commit"""
        self._git(f"checkout -b {branch_name} {commit}")

    def branch_local_name(self, remote_branch_name):
        """Strip remote from the given remote branch"""
        for remote in self.remotes:
            if remote_branch_name.startswith(remote + "/"):
                return remote_branch_name.split("/", 1)[1]

    def checkout_branch(self, branch_name, create=False):
        """Check out the specified branch. If it does not exists and
    the create parameter is set to True, create a new branch.
    If the branch does not exist and create is set to False,
    raise an exception."""
        try:
            self._git(
                "checkout {}",
                branch_name,
                log_error=True,
                reraise_error=True,
                log_success_msg=f"checked out branch {branch_name}",
            )
        except subprocess.CalledProcessError as e:
            if create:
                self.create_and_checkout_branch(branch_name)
            else:
                raise (e)

    def checkout_paths(self, commit, *args):
        for file_path in args:
            self._git(f"checkout {commit} {file_path}")

    def clean(self):
        self._git("clean -fd")

    def clone(self, no_checkout=False, bare=False):

        taf_logger.info("Repo {}: cloning repository", self.repo_name)
        shutil.rmtree(self.repo_path, True)
        os.makedirs(self.repo_path, exist_ok=True)
        if self.repo_urls is None:
            raise Exception("Cannot clone repository. No urls were specified")
        params = ""
        if bare:
            params = "--bare"
        elif no_checkout:
            params = "--no-checkout"
        for url in self.repo_urls:
            try:
                self._git(
                    "clone {} . {}", url, params, log_success_msg="successfully cloned"
                )
            except subprocess.CalledProcessError:
                taf_logger.error(
                    "Repo {}: cannot clone from url {}", self.repo_name, url
                )
            else:
                break

    def create_and_checkout_branch(self, branch_name, raise_error_if_exists=True):
        flag = "-b" if raise_error_if_exists else "-B"
        self._git(
            "checkout {} {}",
            flag,
            branch_name,
            log_success_msg=f"created and checked out branch {branch_name}",
            log_error=True,
            reraise_error=True,
        )

    def create_branch(self, branch_name, commit=None):
        if commit is None:
            commit = ""
        self._git(
            "branch {} {}",
            branch_name,
            commit,
            log_success_msg=f"created branch {branch_name}",
            log_error=True,
            reraise_error=True,
        )

    def create_local_branch(self, branch_name):
        """Create local branch by checking it if it does not exist out and making sure
        to check out previously checked out branch
        """
        if not self.branch_exists(branch_name, include_remotes=False):
            current_branch = self.get_current_branch()
            self.checkout_branch(branch_name)
            self.checkout_branch(current_branch)

    def checkout_commit(self, commit):
        self._git("checkout {}", commit, log_success_msg=f"checked out commit {commit}")

    def commit(self, message):
        """Create a commit with the provided message on the currently checked out branch"""
        self._git("add -A")
        try:
            self._git("diff --cached --exit-code --shortstat")
        except subprocess.CalledProcessError:
            run("git", "-C", self.repo_path, "commit", "--quiet", "-m", message)
        return self._git("rev-parse HEAD")

    def commit_empty(self, message):
        run(
            "git",
            "-C",
            self.repo_path,
            "commit",
            "--quiet",
            "--allow-empty",
            "-m",
            message,
        )
        return self._git("rev-parse HEAD")

    def commits_on_branch_and_not_other(
        self, branch1, branch2, include_branching_commit=False
    ):
        """
        Meant to find commits belonging to a branch which branches off of
        a commit from another branch. For example, to find only commits
        on a speculative branch and not on the master branch.
        """

        taf_logger.debug(
            "Repo {}: finding commits which are on branch {}, but not on branch {}",
            self.repo_name,
            branch1,
            branch2,
        )
        commits = self._git(
            "log {} --not {} --no-merges --format=format:%H", branch1, branch2
        )
        commits = commits.split("\n") if commits else []
        if include_branching_commit:
            branching_commit = self._git("rev-list -n 1 {}~1", commits[-1])
            commits.append(branching_commit)
        taf_logger.debug(
            "Repo {}: found the following commits: {}", self.repo_name, commits
        )
        return commits

    def delete_local_branch(self, branch_name, force=False):
        """Deletes local branch."""
        flag = "-D" if force else "-d"
        self._git(f"branch {flag} {branch_name}")

    def delete_remote_branch(self, branch_name, remote=None):
        if remote is None:
            remote = self.remotes[0]
        self._git(f"push {remote} --delete {branch_name}")

    def get_commits_date(self, commit):
        date = self._git("show -s --format=%at {}", commit)
        return date.split(" ", 1)[0]

    def get_commit_message(self, commit):
        """Returns commit message of the given commit"""
        return self._git("log --format=%B -n 1 {}", commit).strip()

    def get_commit_sha(self, behind_head):
        """Get commit sha of HEAD~{behind_head}"""
        return self._git("rev-parse HEAD~{}", behind_head)

    def get_json(self, commit, path):
        s = self.get_file(commit, path)
        return json.loads(s)

    def get_file(self, commit, path):
        return self._git("show {}:{}", commit, path)

    def get_last_branch_by_committer_date(self):
        """Find the latest branch based on committer date. Should only be used for
        testing purposes"""
        branches = self._git("branch --sort=committerdate").strip().split("\n")
        if not len(branches):
            return None
        return branches[-1]

    def get_remote_url(self):
        try:
            return self._git("config --get remote.origin.url").strip()
        except subprocess.CalledProcessError:
            return None

    def delete_branch(self, branch_name):
        self._git("branch -D {}", branch_name)

    def diff_between_revisions(self, revision1=EMPTY_TREE, revision2="HEAD"):
        return self._git("diff {} {}", revision1, revision2)

    def has_remote(self):
        return bool(self._git("remote"))

    def head_commit_sha(self):
        """Finds sha of the commit to which the current HEAD points"""
        try:
            return self._git("rev-parse HEAD")
        except subprocess.CalledProcessError:
            return None

    def fetch(self, fetch_all=False):
        if fetch_all:
            self._git("fetch --all")
        else:
            self._git("fetch")

    def get_current_branch(self):
        """Return current branch."""
        return self._git("rev-parse --abbrev-ref HEAD").strip()

    def get_merge_base(self, branch1, branch2):
        """Finds the best common ancestor between two branches"""
        return self._git(f"merge-base {branch1} {branch2}")

    def get_tracking_branch(self, branch=""):
        """Returns tracking branch name in format origin/branch-name or None if branch does not
        track remote branch.
        """
        try:
            return self._git(f"rev-parse --abbrev-ref {branch}@{{u}}")
        except subprocess.CalledProcessError:
            return None

    def init_repo(self, bare=False):
        if not os.path.isdir(self.repo_path):
            os.makedirs(self.repo_path, exist_ok=True)
        flag = "--bare" if bare else ""
        self._git(f"init {flag}")
        if self.repo_urls is not None and len(self.repo_urls):
            self._git("remote add origin {}", self.repo_urls[0])

    def list_files_at_revision(self, commit, path=""):
        if path is None:
            path = ""
        file_names = self._git("ls-tree -r --name-only {}", commit)
        list_of_files = []
        if not file_names:
            return list_of_files
        for file_in_repo in file_names.split("\n"):
            if not file_in_repo.startswith(path):
                continue
            file_in_repo = os.path.relpath(file_in_repo, path)
            list_of_files.append(file_in_repo)
        return list_of_files

    def list_commits(self, branch="", **kwargs):
        params = []
        for name, value in kwargs.items():
            params.append(f"--{name}={value}")

        return self._git("log {} {}", branch, " ".join(params)).split("\n")

    def list_commit_shas(self, branch="master"):
        return self.list_commits(branch, format="format:%H")

    def list_n_commits(self, number=10, skip=None, branch=None):
        """List the specified number of top commits of the current branch
        Optionally skip a number of commits from the top"""
        if branch is None:
            branch = ""
        if skip:
            commits = self._git(
                f"log {branch} --format=format:%H --skip={skip} -n {number}"
            )
        else:
            commits = self._git(f"log {branch} --format=format:%H -n {number}")
        return commits.split("\n") if commits else []

    def merge_commit(self, commit):
        self._git("merge {}", commit)

    def pull(self):
        """Pull current branch"""
        self._git("pull")

    def push(self, branch=None, set_upstream=False, force=False):
        """Push all changes"""
        if branch is None:
            branch = self.get_current_branch()
        upstream_flag = "-u" if set_upstream else ""
        force_flag = "-f" if force else ""
        self._git("push {} {} origin {}", upstream_flag, force_flag, branch)

    def rename_branch(self, old_name, new_name):
        self._git("branch -m {} {}", old_name, new_name)

    def reset_num_of_commits(self, num_of_commits, hard=False):
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} HEAD~{num_of_commits}")

    def reset_to_commit(self, commit, hard=False):
        flag = "--hard" if hard else "--soft"
        self._git(f"reset {flag} {commit}")

    def reset_to_head(self):
        self._git("reset --hard HEAD")

    def set_remote_url(self, new_url, remote="origin"):
        self._git(f"remote set-url {remote} {new_url}")

    def set_upstream(self, branch_name):
        self._git("branch -u origin/{}", branch_name)

    def something_to_commit(self):
        """Checks if there are any uncommitted changes"""
        uncommitted_changes = self._git("status --porcelain")
        return bool(uncommitted_changes)

    def synced_with_remote(self, branch="master"):
        """Checks if local branch is synced with its remote branch"""
        # check if the latest local commit matches
        # the latest remote commit on the specified branch
        tracking_branch = self.get_tracking_branch(branch)
        if not tracking_branch:
            return False

        try:
            local_commit = self._git(f"rev-parse {branch}")
        except subprocess.CalledProcessError as e:
            if "unknown revision or path not in the working tree" not in e.output:
                raise e
            local_commit = None

        try:
            remote_commit = self._git(f"rev-parse {tracking_branch}")
        except subprocess.CalledProcessError as e:
            if "unknown revision or path not in the working tree" not in e.output:
                raise e
            remote_commit = None

        return local_commit == remote_commit

    def top_commit_of_branch(self, branch):
        return self._git(f"rev-parse {branch}")


class NamedGitRepository(GitRepository):
    def __init__(
        self,
        root_dir,
        repo_name,
        repo_urls=None,
        additional_info=None,
        default_branch="master",
    ):
        """
    Args:
      root_dir: the root directory
      repo_name: repository's path relative to the root directory root_dir
      repo_urls: repository's urls (optional)
      additional_info: a dictionary containing other data (optional)
      default_branch: repository's default branch
    repo_path is the absolute path to this repository. It is set by joining
    root_dir and repo_name.
    """
        repo_path = _get_repo_path(root_dir, repo_name)
        super().__init__(repo_path, repo_urls, additional_info, default_branch)
        self.repo_name = repo_name


def _get_repo_path(root_dir, repo_name):
    """
  get the path to a repo and ensure it is valid.
  (since this is coming from potentially untrusted data)
  """
    _validate_repo_name(repo_name)
    repo_dir = str((Path(root_dir) / (repo_name or "")))
    if not repo_dir.startswith(repo_dir):
        taf_logger.error("Repo {}: repository name is not valid", repo_name)
        raise InvalidRepositoryError(f"Invalid repository name: {repo_name}")
    return repo_dir


_repo_name_re = re.compile(r"^\w[\w_-]*/\w[\w_-]*$")


def _validate_repo_name(repo_name):
    """ Ensure the repo name is not malicious """
    match = _repo_name_re.match(repo_name)
    if not match:
        taf_logger.error("Repo {}: repository name is not valid", repo_name)
        raise InvalidRepositoryError(
            "Repository name must be in format namespace/repository "
            "and can only contain letters, numbers, underscores and "
            f'dashes, but got "{repo_name}"'
        )


_http_fttp_url = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    # domain...
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

_ssh_url = re.compile(
    r"((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)?(/)?"
)


def _validate_url(url):
    """ ensure valid URL """
    for _url_re in [_http_fttp_url, _ssh_url]:
        match = _url_re.match(url)
        if match:
            return
    taf_logger.error("Repository URL ({}) is not valid", url)
    raise InvalidRepositoryError(
        f'Repository URL must be a valid URL, but got "{url}".'
    )



import os
import sys
from loguru import logger as taf_logger
from pathlib import Path

import taf.settings

_FORMAT_STRING = "[{time}] [{level}] [{module} {function}@{line}]\n{message}\n"


def _get_log_location():
    location = taf.settings.LOGS_LOCATION or os.environ.get("TAF_LOG")
    if location is None:
        location = Path.home() / ".taf"
        location.mkdir(exist_ok=True)
    else:
        location = Path(location)
    return location


taf_logger.remove()

if taf.settings.ENABLE_CONSOLE_LOGGING:
    taf_logger.add(
        sys.stdout, format=_FORMAT_STRING, level=taf.settings.CONSOLE_LOGGING_LEVEL
    )
    taf_logger.add(
        sys.stderr, format=_FORMAT_STRING, level=taf.settings.ERROR_LOGGING_LEVEL
    )


if taf.settings.ENABLE_FILE_LOGGING:
    logs_location = _get_log_location()
    log_path = str(logs_location / taf.settings.LOG_FILENAME)
    taf_logger.add(
        log_path, format=_FORMAT_STRING, level=taf.settings.FILE_LOGGING_LEVEL
    )

    if taf.settings.SEPARATE_ERRORS:
        error_log_path = str(logs_location / taf.settings.ERROR_LOG_FILENAME)
        taf_logger.add(
            error_log_path,
            format=_FORMAT_STRING,
            level=taf.settings.ERROR_LOGGING_LEVEL,
        )



import json
import os
from pathlib import Path

from pytest import fixture

from taf.git import GitRepository


@fixture(autouse=True)
def run_around_tests(taf_happy_path):
    yield
    repo = GitRepository(taf_happy_path.repository_path)
    repo.reset_to_head()
    repo.clean()
    taf_happy_path._repository.targets.clear_targets()
    files_to_keep = []
    for root, _, filenames in os.walk(str(taf_happy_path.targets_path)):
        for filename in filenames:
            relpath = Path(root, filename).relative_to(taf_happy_path.targets_path)
            files_to_keep.append(relpath.as_posix())
    taf_happy_path.add_targets({}, files_to_keep=files_to_keep)


def test_add_targets_new_files(taf_happy_path):

    old_targets = _get_old_targets(taf_happy_path)

    json_file_content = {"attr1": "value1", "attr2": "value2"}
    regular_file_content = "this file is not empty"
    data = {
        "new_json_file": {"target": json_file_content},
        "new_file": {"target": regular_file_content},
        "empty_file": {"target": None},
    }
    taf_happy_path.add_targets(data)
    _check_target_files(taf_happy_path, data, old_targets)


def test_add_targets_nested_files(taf_happy_path):

    old_targets = _get_old_targets(taf_happy_path)

    data = {
        "inner_folder1/new_file_1": {"target": "file 1 content"},
        "inner_folder2/new_file_2": {"target": "file 2 content"},
    }
    taf_happy_path.add_targets(data)
    _check_target_files(taf_happy_path, data, old_targets)


def test_add_targets_files_to_keep(taf_happy_path):
    old_targets = _get_old_targets(taf_happy_path)
    data = {"a_new_file": {"target": "new file content"}}
    files_to_keep = ["branch"]
    taf_happy_path.add_targets(data, files_to_keep=files_to_keep)
    _check_target_files(taf_happy_path, data, old_targets, files_to_keep)


def _check_target_files(repo, data, old_targets, files_to_keep=None):
    if files_to_keep is None:
        files_to_keep = []

    targets_path = repo.targets_path
    for target_rel_path, content in data.items():
        target_path = targets_path / target_rel_path
        assert target_path.exists()
        with open(str(target_path)) as f:
            file_content = f.read()
            target_content = content["target"]
            if isinstance(target_content, dict):
                content_json = json.loads(file_content)
                assert content_json == target_content
            elif target_content:
                assert file_content == target_content
            else:
                assert file_content == ""

    # make sure that everything defined in repositories.json still exists
    repository_targets = []
    repositories_path = targets_path / "repositories.json"
    assert repositories_path.exists()
    with open(str(repositories_path)) as f:
        repositories = json.load(f)["repositories"]
        for target_rel_path in repositories:
            target_path = targets_path / target_rel_path
            assert target_path.exists()
            repository_targets.append(target_rel_path)

    # make sure that files to keep exist
    for file_to_keep in files_to_keep:
        # if the file didn't exists prior to adding new targets
        # it won't exists after adding them
        if file_to_keep not in old_targets:
            continue
        target_path = targets_path / file_to_keep
        assert target_path.exists()

    for old_target in old_targets:
        if (
            old_target not in repository_targets
            and old_target not in data
            and old_target not in repo._framework_files
            and old_target not in files_to_keep
        ):
            assert (targets_path / old_target).exists() is False


def _get_old_targets(repo):
    targets_path = repo.targets_path
    old_targets = []
    for root, _, filenames in os.walk(str(targets_path)):
        for filename in filenames:
            rel_path = Path(root, filename).relative_to(targets_path)
            old_targets.append(rel_path.as_posix())
    return old_targets



import json
import os
from collections import defaultdict
from pathlib import Path
from subprocess import CalledProcessError
from taf.log import taf_logger
from taf.git import GitRepository, NamedGitRepository


class AuthRepoMixin(object):

    LAST_VALIDATED_FILENAME = "last_validated_commit"

    @property
    def conf_dir(self):
        """
        Returns location of the directory which stores the authentication repository's
        configuration files. That is, the last validated commit.
        Create the directory if it does not exist.
        """
        # the repository's name consists of the namespace and name (namespace/name)
        # the configuration directory should be _name
        last_dir = os.path.basename(os.path.normpath(self.repo_path))
        conf_path = Path(self.repo_path).parent / f"_{last_dir}"
        conf_path.mkdir(parents=True, exist_ok=True)
        return str(conf_path)

    @property
    def certs_dir(self):
        certs_dir = Path(self.repo_path, "certs")
        certs_dir.mkdir(parents=True, exist_ok=True)
        return str(certs_dir)

    @property
    def last_validated_commit(self):
        """
        Return the last validated commit of the authentication repository
        """
        try:
            return Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).read_text()
        except FileNotFoundError:
            return None

    def get_target(self, target_name, commit=None, safely=True):
        if commit is None:
            commit = self.head_commit_sha()
        target_path = (Path(self.targets_path) / target_name).as_posix()
        if safely:
            return self._safely_get_json(commit, target_path)
        else:
            return self.get_json(commit, target_path)

    def is_commit_authenticated(self, target_name, commit):
        """Checks if passed commit is ever authenticated for given target name.
        """
        for auth_commit in self.all_commits_on_branch(reverse=False):
            target = self.get_target(target_name, auth_commit)
            try:
                if target["commit"] == commit:
                    return True
            except TypeError:
                continue
        return False

    def set_last_validated_commit(self, commit):
        """
        Set the last validated commit of the authentication repository
        """
        taf_logger.debug(
            "Auth repo {}: setting last validated commit to: {}", self.repo_name, commit
        )
        Path(self.conf_dir, self.LAST_VALIDATED_FILENAME).write_text(commit)

    def sorted_commits_per_repositories(self, commits):
        """Create a list of of subsequent commits per repository
        keeping in mind that targets metadata file is not updated
        everytime something is committed to the authentication repo
        """
        repositories_commits = defaultdict(list)
        targets = self.target_commits_at_revisions(commits)
        previous_commits = {}
        for commit in commits:
            for target_path, target_commit in targets[commit].items():
                previous_commit = previous_commits.get(target_path)
                if previous_commit is None or target_commit != previous_commit:
                    repositories_commits[target_path].append(target_commit)
                previous_commits[target_path] = target_commit
        taf_logger.debug(
            "Auth repo {}: new commits per repositories according to targets.json: {}",
            self.repo_name,
            repositories_commits,
        )
        return repositories_commits

    def target_commits_at_revisions(self, commits):
        targets = defaultdict(dict)
        for commit in commits:
            targets_at_revision = self._safely_get_json(
                commit, self.metadata_path + "/targets.json"
            )
            if targets_at_revision is None:
                continue
            targets_at_revision = targets_at_revision["signed"]["targets"]

            repositories_at_revision = self._safely_get_json(
                commit, self.targets_path + "/repositories.json"
            )
            if repositories_at_revision is None:
                continue
            repositories_at_revision = repositories_at_revision["repositories"]

            for target_path in targets_at_revision:
                if target_path not in repositories_at_revision:
                    # we only care about repositories
                    continue
                try:
                    target_commit = self.get_json(
                        commit, self.targets_path + "/" + target_path
                    ).get("commit")
                    targets[commit][target_path] = target_commit
                except json.decoder.JSONDecodeError:
                    taf_logger.debug(
                        "Auth repo {}: target file {} is not a valid json at revision {}",
                        self.repo_name,
                        target_path,
                        commit,
                    )
                    continue
        return targets

    def _safely_get_json(self, commit, path):
        try:
            return self.get_json(commit, path)
        except CalledProcessError:
            taf_logger.info(
                "Auth repo {}: {} not available at revision {}",
                self.repo_name,
                os.path.basename(path),
                commit,
            )
        except json.decoder.JSONDecodeError:
            taf_logger.info(
                "Auth repo {}: {} not a valid json at revision {}",
                self.repo_name,
                os.path.basename(path),
                commit,
            )


class AuthenticationRepo(AuthRepoMixin, GitRepository):
    def __init__(
        self,
        repo_path,
        metadata_path="metadata",
        targets_path="targets",
        repo_urls=None,
        additional_info=None,
        default_branch="master",
    ):
        super().__init__(repo_path, repo_urls, additional_info, default_branch)
        self.targets_path = targets_path
        self.metadata_path = metadata_path


class NamedAuthenticationRepo(AuthRepoMixin, NamedGitRepository):
    def __init__(
        self,
        root_dir,
        repo_name,
        metadata_path="metadata",
        targets_path="targets",
        repo_urls=None,
        additional_info=None,
        default_branch="master",
    ):

        super().__init__(
            root_dir, repo_name, repo_urls, additional_info, default_branch
        )
        self.targets_path = targets_path
        self.metadata_path = metadata_path



import click
import datetime
from contextlib import contextmanager
from functools import wraps

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from tuf.repository_tool import import_rsakey_from_pem
from ykman.descriptor import list_devices, open_device
from ykman.piv import (
    ALGO,
    DEFAULT_MANAGEMENT_KEY,
    PIN_POLICY,
    SLOT,
    PivController,
    WrongPin,
    generate_random_management_key,
)
from ykman.util import TRANSPORT

from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import YubikeyError, InvalidPINError
from taf.utils import get_pin_for

DEFAULT_PIN = "123456"
DEFAULT_PUK = "12345678"
EXPIRATION_INTERVAL = 36500

_pins_dict = {}


def add_key_pin(serial_num, pin):
    _pins_dict[serial_num] = pin


def get_key_pin(serial_num):
    return _pins_dict.get(serial_num)


def raise_yubikey_err(msg=None):
    """Decorator used to catch all errors raised by yubikey-manager and raise
    YubikeyError. We don't need to handle specific cases.
    """

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except YubikeyError:
                raise
            except Exception as e:
                err_msg = (
                    f"{msg} Reason: ({type(e).__name__}) {str(e)}" if msg else str(e)
                )
                raise YubikeyError(err_msg) from e

        return decorator

    return wrapper


@contextmanager
def _yk_piv_ctrl(serial=None, pub_key_pem=None):
    """Context manager to open connection and instantiate piv controller.

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        - ykman.piv.PivController

    Raises:
        - YubikeyError
    """
    # If pub_key_pem is given, iterate all devices, read x509 certs and try to match
    # public keys.
    if pub_key_pem is not None:
        for yk in list_devices(transports=TRANSPORT.CCID):
            yk_ctrl = PivController(yk.driver)
            device_pub_key_pem = (
                yk_ctrl.read_certificate(SLOT.SIGNATURE)
                .public_key()
                .public_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                )
                .decode("utf-8")
            )
            # Tries to match without last newline char
            if (
                device_pub_key_pem == pub_key_pem
                or device_pub_key_pem[:-1] == pub_key_pem
            ):
                break
            else:
                yk.close()

    else:
        yk = open_device(transports=TRANSPORT.CCID, serial=serial)
        yk_ctrl = PivController(yk.driver)

    yield yk_ctrl, yk.serial
    yk.close()


def is_inserted():
    """Checks if YubiKey is inserted.

    Args:
        None

    Returns:
        True if at least one Yubikey is inserted (bool)

    Raises:
        - YubikeyError
    """
    return len(list(list_devices(transports=TRANSPORT.CCID))) > 0


@raise_yubikey_err()
def is_valid_pin(pin):
    """Checks if given pin is valid.

    Args:
        pin(str): Yubikey piv PIN

    Returns:
        tuple: True if PIN is valid, otherwise False, number of PIN retries

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        try:
            ctrl.verify(pin)
            return True, None  # ctrl.get_pin_tries() fails if PIN is valid
        except WrongPin:
            return False, ctrl.get_pin_tries()


@raise_yubikey_err("Cannot get serial number.")
def get_serial_num(pub_key_pem=None):
    """Get Yubikey serial number.

    Args:
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Yubikey serial number

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (_, serial):
        return serial


@raise_yubikey_err("Cannot export x509 certificate.")
def export_piv_x509(cert_format=serialization.Encoding.PEM, pub_key_pem=None):
    """Exports YubiKey's piv slot x509.

    Args:
        - cert_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV x509 certificate in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        x509 = ctrl.read_certificate(SLOT.SIGNATURE)
        return x509.public_bytes(encoding=cert_format)


@raise_yubikey_err("Cannot export public key.")
def export_piv_pub_key(pub_key_format=serialization.Encoding.PEM, pub_key_pem=None):
    """Exports YubiKey's piv slot public key.

    Args:
        - pub_key_format(str): One of 'serialization.Encoding' formats.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        PIV public key in a given format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        x509 = ctrl.read_certificate(SLOT.SIGNATURE)
        return x509.public_key().public_bytes(
            encoding=pub_key_format,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )


@raise_yubikey_err("Cannot get public key in TUF format.")
def get_piv_public_key_tuf(scheme=DEFAULT_RSA_SIGNATURE_SCHEME, pub_key_pem=None):
    """Return public key from a Yubikey in TUF's RSAKEY_SCHEMA format.

    Args:
        - scheme(str): Rsa signature scheme (default is rsa-pkcs1v15-sha256)
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        A dictionary containing the RSA keys and other identifying information
        from inserted smart card.
        Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.

    Raises:
        - YubikeyError
    """
    pub_key_pem = export_piv_pub_key(pub_key_pem=pub_key_pem).decode("utf-8")
    return import_rsakey_from_pem(pub_key_pem, scheme)


@raise_yubikey_err("Cannot sign data.")
def sign_piv_rsa_pkcs1v15(data, pin, pub_key_pem=None):
    """Sign data with key from YubiKey's piv slot.

    Args:
        - data(bytes): Data to be signed
        - pin(str): Pin for piv slot login.
        - pub_key_pem(str): Match Yubikey's public key (PEM) if multiple keys
                            are inserted

    Returns:
        Signature (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl(pub_key_pem=pub_key_pem) as (ctrl, _):
        ctrl.verify(pin)
        return ctrl.sign(SLOT.SIGNATURE, ALGO.RSA2048, data)


@raise_yubikey_err("Cannot setup Yubikey.")
def setup(
    pin,
    cert_cn,
    cert_exp_days=365,
    pin_retries=10,
    private_key_pem=None,
    mgm_key=generate_random_management_key(),
):
    """Use to setup inserted Yubikey, with following steps (order is important):
      - reset to factory settings
      - set management key
      - generate key(RSA2048) or import given one
      - generate and import self-signed certificate(X509)
      - set pin retries
      - set pin
      - set puk(same as pin)

    Args:
        - cert_cn(str): x509 common name
        - cert_exp_days(int): x509 expiration (in days from now)
        - pin_retries(int): Number of retries for PIN
        - private_key_pem(str): Private key in PEM format. If given, it will be
                                imported to Yubikey.
        - mgm_key(bytes): New management key

    Returns:
        PIV public key in PEM format (bytes)

    Raises:
        - YubikeyError
    """
    with _yk_piv_ctrl() as (ctrl, _):
        # Factory reset and set PINs
        ctrl.reset()

        ctrl.authenticate(DEFAULT_MANAGEMENT_KEY)
        ctrl.set_mgm_key(mgm_key)

        # Generate RSA2048
        if private_key_pem is None:
            pub_key = ctrl.generate_key(SLOT.SIGNATURE, ALGO.RSA2048, PIN_POLICY.ALWAYS)
        else:
            private_key = load_pem_private_key(private_key_pem, None, default_backend())
            ctrl.import_key(SLOT.SIGNATURE, private_key, PIN_POLICY.ALWAYS)
            pub_key = private_key.public_key()

        ctrl.authenticate(mgm_key)
        ctrl.verify(DEFAULT_PIN)

        # Generate and import certificate
        now = datetime.datetime.now()
        valid_to = now + datetime.timedelta(days=cert_exp_days)
        ctrl.generate_self_signed_certificate(
            SLOT.SIGNATURE, pub_key, cert_cn, now, valid_to
        )

        ctrl.set_pin_retries(pin_retries=pin_retries, puk_retries=pin_retries)
        ctrl.change_pin(DEFAULT_PIN, pin)
        ctrl.change_puk(DEFAULT_PUK, pin)

    return pub_key.public_bytes(
        serialization.Encoding.PEM, serialization.PublicFormat.SubjectPublicKeyInfo
    )


def setup_new_yubikey(serial_num, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
    pin = get_key_pin(serial_num)
    cert_cn = input("Enter key holder's name: ")
    print("Generating key, please wait...")
    pub_key_pem = setup(pin, cert_cn, cert_exp_days=EXPIRATION_INTERVAL).decode("utf-8")
    scheme = DEFAULT_RSA_SIGNATURE_SCHEME
    key = import_rsakey_from_pem(pub_key_pem, scheme)
    return key


def get_and_validate_pin(key_name, pin_confirm=True, pin_repeat=True):
    valid_pin = False
    while not valid_pin:
        pin = get_pin_for(key_name, pin_confirm, pin_repeat)
        valid_pin, retries = is_valid_pin(pin)
        if not valid_pin and not retries:
            raise InvalidPINError("No retries left. YubiKey locked.")
        if not valid_pin:
            if not click.confirm(
                f"Incorrect PIN. Do you want to try again? {retries} retires left."
            ):
                raise InvalidPINError("PIN input cancelled")
    return pin


def yubikey_prompt(
    key_name,
    role=None,
    taf_repo=None,
    registering_new_key=False,
    creating_new_key=False,
    loaded_yubikeys=None,
    pin_confirm=True,
    pin_repeat=True,
    prompt_message=None,
):
    def _read_and_check_yubikey(
        key_name,
        role,
        taf_repo,
        registering_new_key,
        creating_new_key,
        loaded_yubikeys,
        pin_confirm,
        pin_repeat,
        prompt_message,
    ):

        if prompt_message is None:
            prompt_message = f"Please insert {key_name} YubiKey and press ENTER"
        input(prompt_message)
        # make sure that YubiKey is inserted
        try:
            serial_num = get_serial_num()
        except Exception:
            print("YubiKey not inserted")
            return False, None, None

        # check if this key is already loaded as the provided role's key (we can use the same key
        # to sign different metadata)
        if (
            loaded_yubikeys is not None
            and serial_num in loaded_yubikeys
            and role in loaded_yubikeys[serial_num]
        ):
            print("Key already loaded")
            return False, None, None

        # read the public key, unless a new key needs to be generated on the yubikey
        public_key = get_piv_public_key_tuf() if not creating_new_key else None
        # check if this yubikey is can be used for signing the provided role's metadata
        # if the key was already registered as that role's key
        if not registering_new_key and role is not None and taf_repo is not None:
            if not taf_repo.is_valid_metadata_yubikey(role, public_key):
                print(f"The inserted YubiKey is not a valid {role} key")
                return False, None, None

        if get_key_pin(serial_num) is None:
            if creating_new_key:
                pin = get_pin_for(key_name, pin_confirm, pin_repeat)
            else:
                pin = get_and_validate_pin(key_name, pin_confirm, pin_repeat)
            add_key_pin(serial_num, pin)

        if role is not None:
            if loaded_yubikeys is None:
                loaded_yubikeys = {serial_num: [role]}
            else:
                loaded_yubikeys.setdefault(serial_num, []).append(role)

        return True, public_key, serial_num

    while True:
        success, key, serial_num = _read_and_check_yubikey(
            key_name,
            role,
            taf_repo,
            registering_new_key,
            creating_new_key,
            loaded_yubikeys,
            pin_confirm,
            pin_repeat,
            prompt_message,
        )
        if success:
            return key, serial_num



import os
import shutil

import tuf
import tuf.client.updater as tuf_updater

from taf.log import taf_logger
import taf.repositoriesdb as repositoriesdb
import taf.settings as settings
from taf.exceptions import UpdateFailedError
from taf.updater.handlers import GitUpdater
from taf.utils import on_rm_error


def update_repository(
    url,
    clients_repo_path,
    targets_dir,
    update_from_filesystem,
    authenticate_test_repo=False,
    target_repo_classes=None,
    target_factory=None,
):
    """
    <Arguments>
    url:
        URL of the remote authentication repository
    clients_repo_path:
        Client's authentication repository's full path
    targets_dir:
        Directory where the target repositories are located
    update_from_filesystem:
        A flag which indicates if the URL is actually a file system path
    authenticate_test_repo:
        A flag which indicates that the repository to be updated is a test repository
    target_repo_classes:
        A class or a dictionary used when instantiating target repositories.
        See repositoriesdb load_repositories for more details.
    target_factory:
        A git repositories factory used when instantiating target repositories.
        See repositoriesdb load_repositories for more details.
    """
    # if the repository's name is not provided, divide it in parent directory
    # and repository name, since TUF's updater expects a name
    # but set the validate_repo_name setting to False
    clients_dir, repo_name = os.path.split(os.path.normpath(clients_repo_path))
    settings.validate_repo_name = False
    update_named_repository(
        url,
        clients_dir,
        repo_name,
        targets_dir,
        update_from_filesystem,
        authenticate_test_repo,
        target_repo_classes,
        target_factory,
    )


def update_named_repository(
    url,
    clients_directory,
    repo_name,
    targets_dir,
    update_from_filesystem,
    authenticate_test_repo=False,
    target_repo_classes=None,
    target_factory=None,
):
    """
    <Arguments>
        url:
        URL of the remote authentication repository
        clients_directory:
        Directory where the client's authentication repository is located
        repo_name:
        Name of the authentication repository. Can be namespace prefixed
        targets_dir:
        Directory where the target repositories are located
        update_from_filesystem:
        A flag which indicates if the URL is actually a file system path
        authenticate_test_repo:
        A flag which indicates that the repository to be updated is a test repository
        target_repo_classes:
        A class or a dictionary used when instantiating target repositories.
        See repositoriesdb load_repositories for more details.
        target_factory:
        A git repositories factory used when instantiating target repositories.
        See repositoriesdb load_repositories for more details.

    The general idea of the updater is the following:
    - We have a git repository which contains the metadata files. These metadata files
    are in the 'metadata' directory
    - Clients have a clone of that repository on their local machine and want to update it
    - We don't want to simply pull the updates. We want to verify that the new commits
    (committed after the most recent one in the client's local repository)
    - For each of the new commits, we want to check if all metadata is valid. The set of
    metadata should be valid as a whole at that revision. Not only do we want to make sure
    that a metadata which is supposed to be changed was indeed updated and is valid, but
    also to make sure that if a metadata file should not be updated, it remained the same.
    - We also want to make sure that all targets metadata is valid (including the delegated roles)
    - We do not want to simply update the metadata to the latest version, without skipping
    these checks. We want to check each commit, not just the last one.
    - If we are checking a commit which is not the latest one, we do not want to report an error
    if the metadata expired. We want to make sure that that was valid at the time when the
    metadata was committed.
    - We can rely on the TUF's way of handling metadata, by using the current and previous
    directories. We just want to automatically create and update them. They should not
    remain on the client's machine.
    - We do not want to modify TUF's updater to much, but still need to get around the fact
    that TUF skips mirrors which do not have valid and/or current metadata files. Also, we
    do not simply want to find the latest metadata, we want to validate everything in-between.
    That is why the idea is to call refresh multiple times, until the last commit is reached.
    The 'GitMetadataUpdater' updater is designed in such a way that for each new call it
    loads data from a most recent commit.
    """

    # at the moment, we assume that the initial commit is valid and that it contains at least root.json

    settings.update_from_filesystem = update_from_filesystem
    # instantiate TUF's updater
    repository_mirrors = {
        "mirror1": {
            "url_prefix": url,
            "metadata_path": "metadata",
            "targets_path": "targets",
            "confined_target_dirs": [""],
        }
    }

    tuf.settings.repositories_directory = clients_directory
    repository_updater = tuf_updater.Updater(repo_name, repository_mirrors, GitUpdater)
    users_auth_repo = repository_updater.update_handler.users_auth_repo
    existing_repo = users_auth_repo.is_git_repository_root
    try:
        validation_auth_repo = repository_updater.update_handler.validation_auth_repo
        commits = repository_updater.update_handler.commits
        last_validated_commit = users_auth_repo.last_validated_commit
        if last_validated_commit is None:
            # check if the repository being updated is a test repository
            targets = validation_auth_repo.get_json(
                commits[-1], "metadata/targets.json"
            )
            test_repo = "test-auth-repo" in targets["signed"]["targets"]
            if test_repo and not authenticate_test_repo:
                raise UpdateFailedError(
                    f"Repository {users_auth_repo.repo_name} is a test repository."
                    'Call update with "--authenticate-test-repo to update a test "'
                    "repository"
                )
            elif not test_repo and authenticate_test_repo:
                raise UpdateFailedError(
                    f"Repository {users_auth_repo.repo_name} is not a test repository,"
                    ' but update was called with the "--authenticate-test-repo" flag'
                )

        # validate the authentication repository and fetch new commits
        _update_authentication_repository(repository_updater)

        # get target repositories and their commits, as specified in targets.json

        repositoriesdb.load_repositories(
            users_auth_repo,
            repo_classes=target_repo_classes,
            factory=target_factory,
            root_dir=targets_dir,
            commits=commits,
        )
        repositories = repositoriesdb.get_deduplicated_repositories(
            users_auth_repo, commits
        )
        repositories_commits = users_auth_repo.sorted_commits_per_repositories(commits)

        # update target repositories
        repositories_json = users_auth_repo.get_json(
            commits[-1], "targets/repositories.json"
        )
        last_validated_commit = users_auth_repo.last_validated_commit
        _update_target_repositories(
            repositories, repositories_json, repositories_commits, last_validated_commit
        )

        last_commit = commits[-1]
        taf_logger.info(
            "Merging commit {} into {}", last_commit, users_auth_repo.repo_name
        )
        # if there were no errors, merge the last validated authentication repository commit
        users_auth_repo.checkout_branch(users_auth_repo.default_branch)
        users_auth_repo.merge_commit(last_commit)
        # update the last validated commit
        users_auth_repo.set_last_validated_commit(last_commit)
    except Exception as e:
        if not existing_repo:
            shutil.rmtree(users_auth_repo.repo_path, onerror=on_rm_error)
            shutil.rmtree(users_auth_repo.conf_dir)
        raise e
    finally:
        repositoriesdb.clear_repositories_db()


def _update_authentication_repository(repository_updater):

    users_auth_repo = repository_updater.update_handler.users_auth_repo
    taf_logger.info(
        "Validating authentication repository {}", users_auth_repo.repo_name
    )
    try:
        while not repository_updater.update_handler.update_done():
            current_commit = repository_updater.update_handler.current_commit
            repository_updater.refresh()
            # using refresh, we have updated all main roles
            # we still need to update the delegated roles (if there are any)
            # that is handled by get_current_targets
            current_targets = repository_updater.update_handler.get_current_targets()
            taf_logger.debug("Validated metadata files at revision {}", current_commit)
            for target_path in current_targets:
                target = repository_updater.get_one_valid_targetinfo(target_path)
                target_filepath = target["filepath"]
                trusted_length = target["fileinfo"]["length"]
                trusted_hashes = target["fileinfo"]["hashes"]
                try:
                    repository_updater._get_target_file(
                        target_filepath, trusted_length, trusted_hashes
                    )  # pylint: disable=W0212 # noqa
                except tuf.exceptions.NoWorkingMirrorError as e:
                    taf_logger.error("Could not validate file {}", target_filepath)
                    raise e
                taf_logger.debug(
                    "Successfully validated target file {} at {}",
                    target_filepath,
                    current_commit,
                )
    except Exception as e:
        # for now, useful for debugging
        taf_logger.error(
            "Validation of authentication repository {} failed due to error {}",
            users_auth_repo.repo_name,
            e,
        )
        raise UpdateFailedError(
            f"Validation of authentication repository {users_auth_repo.repo_name}"
            f" failed due to error: {e}"
        )
    finally:
        repository_updater.update_handler.cleanup()

    taf_logger.info(
        "Successfully validated authentication repository {}", users_auth_repo.repo_name
    )
    # fetch the latest commit or clone the repository without checkout
    # do not merge before targets are validated as well
    if users_auth_repo.is_git_repository_root:
        users_auth_repo.fetch(True)
    else:
        users_auth_repo.clone(no_checkout=True)


def _update_target_repositories(
    repositories, repositories_json, repositories_commits, last_validated_commit
):
    taf_logger.info("Validating target repositories")

    # keep track of the repositories which were cloned
    # so that they can be removed if the update fails
    cloned_repositories = []
    allow_unauthenticated = {}
    new_commits = {}

    for path, repository in repositories.items():

        allow_unauthenticated_for_repo = (
            repositories_json["repositories"][repository.repo_name]
            .get("custom", {})
            .get("allow-unauthenticated-commits", False)
        )
        allow_unauthenticated[path] = allow_unauthenticated_for_repo

        # if last_validated_commit is None, start the update from the beginning

        is_git_repository = repository.is_git_repository_root
        if last_validated_commit is None or not is_git_repository:
            old_head = None
        else:
            old_head = repository.head_commit_sha()

        if old_head is None and not is_git_repository:
            repository.clone(no_checkout=True)
            cloned_repositories.append(repository)
        else:
            repository.fetch(True)

        if old_head is not None:
            if allow_unauthenticated:
                old_head = repositories_commits[path][0]
            new_commits_for_repo = repository.all_fetched_commits()
            new_commits_for_repo.insert(0, old_head)
        else:
            new_commits_for_repo = repository.all_commits_since_commit(old_head)
            if is_git_repository:
                # this happens in the case when last_validated_commit does not exist
                # we want to validate all commits, so combine existing commits and
                # fetched commits
                fetched_commits = repository.all_fetched_commits()
                new_commits_for_repo.extend(fetched_commits)
        new_commits[path] = new_commits_for_repo

        try:
            _update_target_repository(
                repository,
                new_commits_for_repo,
                repositories_commits[path],
                allow_unauthenticated_for_repo,
            )
        except UpdateFailedError as e:
            # delete all repositories that were cloned
            for repo in cloned_repositories:
                taf_logger.debug("Removing cloned repository {}", repo.repo_path)
                shutil.rmtree(repo.repo_path, onerror=on_rm_error)
            # TODO is it important to undo a fetch if the repository was not cloned?
            raise e

    taf_logger.info("Successfully validated all target repositories.")
    # if update is successful, merge the commits
    for path, repository in repositories.items():
        repository.checkout_branch(repository.default_branch)
        if len(repositories_commits[path]):
            taf_logger.info(
                "Merging {} into {}",
                repositories_commits[path][-1],
                repository.repo_name,
            )
            last_validated_commit = repositories_commits[path][-1]
            commit_to_merge = (
                last_validated_commit
                if not allow_unauthenticated[path]
                else new_commits[path][-1]
            )
            repository.merge_commit(commit_to_merge)
            if not allow_unauthenticated[path]:
                repository.checkout_commit(commit_to_merge)
            else:
                repository.checkout_branch(repository.default_branch)


def _update_target_repository(
    repository, new_commits, target_commits, allow_unauthenticated
):

    taf_logger.info("Validating target repository {}", repository.repo_name)
    # A new commit might have been pushed after the update process
    # started and before fetch was called
    # So, the number of new commits, pushed to the target repository, could
    # be greater than the number of these commits according to the authentication
    # repository. The opposite cannot be the case.
    # In general, if there are additional commits in the target repositories,
    # the updater will finish the update successfully, but will only update the
    # target repositories until the latest validate commit
    if not allow_unauthenticated:
        update_successful = len(new_commits) >= len(target_commits)
        if update_successful:
            for target_commit, repo_commit in zip(target_commits, new_commits):
                if target_commit != repo_commit:
                    update_successful = False
                    break
        if len(new_commits) > len(target_commits):
            additional_commits = new_commits[len(target_commits) :]
            taf_logger.warning(
                "Found commits {} in repository {} that are not accounted for in the authentication repo."
                "Repository will be updated up to commit {}",
                additional_commits,
                repository.repo_name,
                target_commits[-1],
            )
    else:
        taf_logger.info(
            "Unauthenticated commits allowed in repository {}", repository.repo_name
        )
        update_successful = True
        target_commits_index = 0
        for commit in new_commits:
            if commit in target_commits:
                if commit != target_commits[target_commits_index]:
                    update_successful = False
                    break
                else:
                    target_commits_index += 1

        update_successful = target_commits_index == len(target_commits)

    if not update_successful:
        taf_logger.error(
            "Mismatch between target commits specified in authentication repository and the "
            "target repository {}",
            repository.repo_name,
        )
        raise UpdateFailedError(
            "Mismatch between target commits specified in authentication repository"
            f" and target repository {repository.repo_name}"
        )
    taf_logger.info("Successfully validated {}", repository.repo_name)



import datetime
import json
from functools import partial
from pathlib import Path

import securesystemslib
import tuf.repository_tool
from securesystemslib.exceptions import Error as SSLibError
from securesystemslib.interface import import_rsa_privatekey_from_file
from taf.constants import DEFAULT_RSA_SIGNATURE_SCHEME
from taf.exceptions import (
    InvalidKeyError,
    MetadataUpdateError,
    RootMetadataUpdateError,
    SnapshotMetadataUpdateError,
    TargetsMetadataUpdateError,
    TimestampMetadataUpdateError,
    YubikeyError,
)
from taf.git import GitRepository
from taf.utils import normalize_file_line_endings
from tuf.exceptions import Error as TUFError
from tuf.repository_tool import (
    METADATA_DIRECTORY_NAME,
    TARGETS_DIRECTORY_NAME,
    import_rsakey_from_pem,
    load_repository,
)

# Default expiration intervals per role
expiration_intervals = {"root": 365, "targets": 90, "snapshot": 7, "timestamp": 1}

# Loaded keys cache
role_keys_cache = {}


DISABLE_KEYS_CACHING = False


def load_role_key(keystore, role, password=None, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
    """Loads the specified role's key from a keystore file.
    The keystore file can, but doesn't have to be password protected.

    NOTE: Keys inside keystore should match a role name!

    Args:
        - keystore(str): Path to the keystore directory
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - password(str): (Optional) password used for PEM decryption
        - scheme(str): A signature scheme used for signing.

    Returns:
        - An RSA key object, conformant to 'securesystemslib.RSAKEY_SCHEMA'.

    Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.CryptoError: If path is not a valid encrypted key file.
    """
    if not password:
        password = None
    key = role_keys_cache.get(role)
    if key is None:
        if password is not None:
            key = import_rsa_privatekey_from_file(
                str(Path(keystore, role)), password, scheme=scheme
            )
        else:
            key = import_rsa_privatekey_from_file(
                str(Path(keystore, role)), scheme=scheme
            )
    if not DISABLE_KEYS_CACHING:
        role_keys_cache[role] = key
    return key


def targets_signature_provider(key_id, key_pin, key, data):  # pylint: disable=W0613
    """Targets signature provider used to sign data with YubiKey.

    Args:
        - key_id(str): Key id from targets metadata file
        - key_pin(str): PIN for signing key
        - key(securesystemslib.formats.RSAKEY_SCHEMA): Key info
        - data(dict): Data to sign

    Returns:
        Dictionary that comforms to `securesystemslib.formats.SIGNATURE_SCHEMA`

    Raises:
        - YubikeyError: If signing with YubiKey cannot be performed
    """
    from taf.yubikey import sign_piv_rsa_pkcs1v15
    from binascii import hexlify

    data = securesystemslib.formats.encode_canonical(data).encode("utf-8")
    signature = sign_piv_rsa_pkcs1v15(data, key_pin)

    return {"keyid": key_id, "sig": hexlify(signature).decode()}


def root_signature_provider(signature_dict, key_id, _key, _data):
    """Root signature provider used to return signatures created remotely.

    Args:
        - signature_dict(dict): Dict where key is key_id and value is signature
        - key_id(str): Key id from targets metadata file
        - _key(securesystemslib.formats.RSAKEY_SCHEMA): Key info
        - _data(dict): Data to sign (already signed remotely)

    Returns:
        Dictionary that comforms to `securesystemslib.formats.SIGNATURE_SCHEMA`

    Raises:
        - KeyError: If signature for key_id is not present in signature_dict
    """
    from binascii import hexlify

    return {"keyid": key_id, "sig": hexlify(signature_dict.get(key_id)).decode()}


def yubikey_signature_provider(name, key_id, key, data):  # pylint: disable=W0613
    """
    A signatures provider which asks the user to insert a yubikey
    Useful if several yubikeys need to be used at the same time
    """
    import taf.yubikey as yk
    from binascii import hexlify

    data = securesystemslib.formats.encode_canonical(data).encode("utf-8")

    def _check_key_and_get_pin(expected_key_id):
        try:
            input(f"Insert {name} and press enter")
            inserted_key = yk.get_piv_public_key_tuf()
            if expected_key_id != inserted_key["keyid"]:
                return None
            serial_num = yk.get_serial_num(inserted_key)
            pin = yk.get_key_pin(serial_num)
            if pin is None:
                pin = yk.get_and_validate_pin(name)
            return pin
        except Exception:
            return None

    while True:
        pin = _check_key_and_get_pin(key_id)
        if pin is not None:
            break

    signature = yk.sign_piv_rsa_pkcs1v15(data, pin)
    return {"keyid": key_id, "sig": hexlify(signature).decode()}


class Repository:
    def __init__(self, repository_path):
        self.repository_path = repository_path
        tuf.repository_tool.METADATA_STAGED_DIRECTORY_NAME = METADATA_DIRECTORY_NAME
        tuf_repository = load_repository(repository_path)
        self._repository = tuf_repository

    _framework_files = ["repositories.json", "test-auth-repo"]

    @property
    def targets_path(self):
        return Path(self.repository_path, TARGETS_DIRECTORY_NAME)

    @property
    def metadata_path(self):
        return Path(self.repository_path, METADATA_DIRECTORY_NAME)

    @property
    def repo_id(self):
        return GitRepository(self.repository_path).initial_commit

    @property
    def certs_dir(self):
        certs_dir = Path(self.repository_path, "certs")
        certs_dir.mkdir(parents=True, exist_ok=True)
        return str(certs_dir)

    def _add_target(self, targets_obj, file_path, custom=None):
        """
        <Purpose>
        Normalizes line endings (converts all line endings to unix style endings) and
        registers the target file as a TUF target
        <Arguments>
        targets_obj: TUF targets objects (instance of TUF's targets role class)
        file_path: full path of the target file
        custom: custom target data
        """
        normalize_file_line_endings(file_path)
        targets_obj.add_target(file_path, custom)

    def _role_obj(self, role):
        """Helper function for getting TUF's role object, given the role's name

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
        One of metadata objects:
            Root, Snapshot, Timestamp, Targets or delegated metadata

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        """
        if role == "targets":
            return self._repository.targets
        elif role == "snapshot":
            return self._repository.snapshot
        elif role == "timestamp":
            return self._repository.timestamp
        elif role == "root":
            return self._repository.root
        return self._repository.targets(role)

    def _try_load_metadata_key(self, role, key):
        """Check if given key can be used to sign given role and load it.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - key(securesystemslib.formats.RSAKEY_SCHEMA): Private key used to sign metadata

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        - InvalidKeyError: If metadata cannot be signed with given key.
        """
        if not self.is_valid_metadata_key(role, key):
            raise InvalidKeyError(role)
        self._role_obj(role).load_signing_key(key)

    def _update_metadata(
        self, role, start_date=datetime.datetime.now(), interval=None, write=False
    ):
        """Update metadata expiration date and (optionally) writes it.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - start_date(datetime): Date to which the specified interval is added when
                                calculating expiration date. If a value is not
                                provided, it is set to the current time
        - interval(int): Number of days added to the start date. If not provided,
                        the default value is used
        - write(bool): If True metadata will be signed and written

        Returns:
        - None

        Raises:
        - securesystemslib.exceptions.Error: If securesystemslib error happened during metadata write
        - tuf.exceptions.Error: If TUF error happened during metadata write
        """
        self.set_metadata_expiration_date(role, start_date, interval)
        if write:
            self._repository.write(role)

    def add_existing_target(self, file_path, targets_role="targets", custom=None):
        """Registers new target files with TUF.
        The files are expected to be inside the targets directory.

        Args:
        - file_path(str): Path to target file
        - targets_role(str): Targets or delegated role: a targets role (the root targets role
                            or one of the delegated ones)
        - custom(dict): Custom information for given file

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If 'filepath' is improperly formatted.
        - securesystemslib.exceptions.Error: If 'filepath' is not located in the repository's targets
                                            directory.
        """
        targets_obj = self._role_obj(targets_role)
        self._add_target(targets_obj, file_path, custom)

    def add_targets(self, data, targets_role="targets", files_to_keep=None):
        """Creates a target.json file containing a repository's commit for each repository.
        Adds those files to the tuf repository. Also removes all targets from the filesystem if their
        path is not among the provided ones. TUF does not delete targets automatically.

        Args:
        - data(dict): Dictionary whose keys are target paths of repositories
                        (as specified in targets.json, relative to the targets dictionary).
                        The values are of form:
                        {
                        target: content of the target file
                        custom: {
                            custom_field1: custom_value1,
                            custom_field2: custom_value2
                        }
                        }

        Content of the target file can be a dictionary, in which case a jason file will be created.
        If that is not the case, an ordinary textual file will be created.
        If content is not specified and the file already exists, it will not be modified.
        If it does not exist, an empty file will be created. To replace an existing file with an
        empty file, specify empty content (target: '')

        If a target exists on disk, but is not specified in the provided targets data dictionary,
        it will be:
        - removed, if the target does not correspond to a repository defined in repositories.json
        - left as is,  if the target does correspondw to a repository defined in repositories.json

        Custom is an optional property which, if present, will be used to specify a TUF target's

        - targets_role(str): Targets or delegated role: a targets role (the root targets role
                            or one of the delegated ones)
        - files_to_keep(list|tuple): List of files defined in the previous version of targets.json
                                    that should remain targets. Files required by the framework will
                                    also remain targets.
        """
        if files_to_keep is None:
            files_to_keep = []
        # leave all files required by the framework and additional files specified by the user
        files_to_keep.extend(self._framework_files)
        # add all repositories defined in repositories.json to files_to_keep
        files_to_keep.extend(self._get_target_repositories())
        # delete files if they no longer correspond to a target defined
        # in targets metadata and are not specified in files_to_keep
        targets_obj = self._role_obj(targets_role)
        for filepath in self.targets_path.rglob("*"):
            if filepath.is_file():
                file_rel_path = str(
                    Path(filepath).relative_to(self.targets_path).as_posix()
                )
                if file_rel_path not in data and file_rel_path not in files_to_keep:
                    if file_rel_path in targets_obj.target_files:
                        targets_obj.remove_target(file_rel_path)
                    filepath.unlink()

        for path, target_data in data.items():
            # if the target's parent directory should not be "targets", create
            # its parent directories if they do not exist
            target_path = (self.targets_path / path).absolute()
            target_dir = target_path.parents[0]
            target_dir.mkdir(parents=True, exist_ok=True)

            # create the target file
            content = target_data.get("target", None)
            if content is None:
                if not target_path.is_file():
                    target_path.touch()
            else:
                with open(str(target_path), "w") as f:
                    if isinstance(content, dict):
                        json.dump(content, f, indent=4)
                    else:
                        f.write(content)

            custom = target_data.get("custom", None)
            self._add_target(targets_obj, str(target_path), custom)

        previous_targets = json.loads(
            Path(self.metadata_path, f"{targets_role}.json").read_text()
        )["signed"]["targets"]

        for path in files_to_keep:
            # if path if both in data and files_to_keep, skip it
            # e.g. repositories.json will always be in files_to_keep,
            # but it might also be specified in data, if it needs to be updated
            if path in data:
                continue
            target_path = (self.targets_path / path).absolute()
            previous_custom = None
            if path in previous_targets:
                previous_custom = previous_targets[path].get("custom")
            if target_path.is_file():
                self._add_target(targets_obj, str(target_path), previous_custom)

    def _get_target_repositories(self):
        repositories_path = self.targets_path / "repositories.json"
        if repositories_path.exists():
            repositories = repositories_path.read_text()
            repositories = json.loads(repositories)["repositories"]
            return [str(Path(target_path).as_posix()) for target_path in repositories]

    def get_role_keys(self, role):
        """Registers new target files with TUF.
        The files are expected to be inside the targets directory.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
        List of the role's keyids (i.e., keyids of the keys).

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        """
        role_obj = self._role_obj(role)
        return role_obj.keys

    def get_role_threshold(self, role):
        return self._role_obj(role).threshold

    def get_signable_metadata(self, role):
        """Return signable portion of newly generate metadata for given role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)

        Returns:
        A string representing the 'object' encoded in canonical JSON form or None

        Raises:
        None
        """
        try:
            from tuf.keydb import get_key

            signable = None

            role_obj = self._role_obj(role)
            key = get_key(role_obj.keys[0])

            def _provider(key, data):
                nonlocal signable
                signable = securesystemslib.formats.encode_canonical(data)

            role_obj.add_external_signature_provider(key, _provider)
            self.writeall()
        except (IndexError, TUFError, SSLibError):
            return signable

    def is_valid_metadata_key(self, role, key):
        """Checks if metadata role contains key id of provided key.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - key(securesystemslib.formats.RSAKEY_SCHEMA): Timestamp key.

        Returns:
        Boolean. True if key id is in metadata role key ids, False otherwise.

        Raises:
        - securesystemslib.exceptions.FormatError: If key does not match RSAKEY_SCHEMA
        - securesystemslib.exceptions.UnknownRoleError: If role does not exist
        """
        securesystemslib.formats.RSAKEY_SCHEMA.check_match(key)

        return key["keyid"] in self.get_role_keys(role)

    def is_valid_metadata_yubikey(self, role, public_key=None):
        """Checks if metadata role contains key id from YubiKey.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one
        - public_key(securesystemslib.formats.RSAKEY_SCHEMA): RSA public key dict

        Returns:
        Boolean. True if smart card key id belongs to metadata role key ids

        Raises:
        - YubikeyError
        - securesystemslib.exceptions.FormatError: If 'PEM' is improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If role does not exist
        """
        securesystemslib.formats.ROLENAME_SCHEMA.check_match(role)

        if public_key is None:
            from taf.yubikey import get_piv_public_key_tuf

            public_key = get_piv_public_key_tuf()

        return self.is_valid_metadata_key(role, public_key)

    def add_metadata_key(self, role, pub_key_pem, scheme=DEFAULT_RSA_SIGNATURE_SCHEME):
        """Add metadata key of the provided role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - pub_key_pem(str|bytes): Public key in PEM format

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        - securesystemslib.exceptions.UnknownKeyError: If 'key_id' is not found in the keydb database.

        """
        if isinstance(pub_key_pem, bytes):
            pub_key_pem = pub_key_pem.decode("utf-8")

        key = import_rsakey_from_pem(pub_key_pem, scheme)
        self._role_obj(role).add_verification_key(key)

    def remove_metadata_key(self, role, key_id):
        """Remove metadata key of the provided role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - key_id(str): An object conformant to 'securesystemslib.formats.KEYID_SCHEMA'.

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        - securesystemslib.exceptions.UnknownKeyError: If 'key_id' is not found in the keydb database.

        """
        from tuf.keydb import get_key

        key = get_key(key_id)
        self._role_obj(role).remove_verification_key(key)

    def set_metadata_expiration_date(
        self, role, start_date=datetime.datetime.now(), interval=None
    ):
        """Set expiration date of the provided role.

        Args:
        - role(str): TUF role (root, targets, timestamp, snapshot or delegated one)
        - start_date(datetime): Date to which the specified interval is added when calculating
                                expiration date. If a value is not provided, it is set to the
                                current time.
        - interval(int): A number of days added to the start date.
                        If not provided, the default value is set based on the role:

                            root - 365 days
                            targets - 90 days
                            snapshot - 7 days
                            timestamp - 1 day
                            all other roles - 1 day

        Returns:
        None

        Raises:
        - securesystemslib.exceptions.FormatError: If the arguments are improperly formatted.
        - securesystemslib.exceptions.UnknownRoleError: If 'rolename' has not been delegated by this
                                                        targets object.
        """
        role_obj = self._role_obj(role)
        if interval is None:
            interval = expiration_intervals.get(role, 1)
        expiration_date = start_date + datetime.timedelta(interval)
        role_obj.expiration = expiration_date

    def get_expiration_date(self, role):
        return self._role_obj(role).expiration

    def update_root(self, signature_dict):
        """Update root metadata.

        Args:
        - signature_dict(dict): key_id-signature dictionary

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - SnapshotMetadataUpdateError: If any other error happened during metadata update
        """
        from tuf.keydb import get_key

        try:
            for key_id in signature_dict:
                key = get_key(key_id)
                self._repository.root.add_external_signature_provider(
                    key, partial(root_signature_provider, signature_dict, key_id)
                )
            self.writeall()
        except (TUFError, SSLibError) as e:
            raise RootMetadataUpdateError(str(e))

    def update_snapshot(
        self,
        snapshot_key,
        start_date=datetime.datetime.now(),
        interval=None,
        write=True,
    ):
        """Update snapshot metadata.

        Args:
        - snapshot_key
        - password: Keystore file password
        - start_date(datetime): Date to which the specified interval is added when
                                calculating expiration date. If a value is not
                                provided, it is set to the current time
        - interval(int): Number of days added to the start date. If not provided,
                        the default value is used
        - write(bool): If True snapshot metadata will be signed and written

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - SnapshotMetadataUpdateError: If any other error happened during metadata update
        """
        try:
            self._try_load_metadata_key("snapshot", snapshot_key)
            self._update_metadata("snapshot", start_date, interval, write=write)
        except (TUFError, SSLibError) as e:
            raise SnapshotMetadataUpdateError(str(e))

    def update_snapshot_and_timestmap(
        self, snapshot_key, timestamp_key, write=True, **kwargs
    ):
        """Update snapshot and timestamp metadata.

        Args:
        - snapshot_key(str): snapshot key
        - timestamp_key(str): timestamp key
        - write(bool): (Optional) If True snapshot and timestamp metadata will be signed and written
        - kwargs(dict): (Optional) Expiration dates and intervals:
                        - snapshot_date(datetime)
                        - snapshot_interval(int)
                        - timestamp_date(datetime)
                        - timestamp_interval(int)

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - MetadataUpdateError: If any other error happened during metadata update
        """
        try:
            snapshot_date = kwargs.get("snapshot_date", datetime.datetime.now())
            snapshot_interval = kwargs.get("snapshot_interval", None)

            timestamp_date = kwargs.get("timestamp_date", datetime.datetime.now())
            timestamp_interval = kwargs.get("timestamp_interval", None)

            self.update_snapshot(
                snapshot_key, snapshot_date, snapshot_interval, write=write
            )
            self.update_timestamp(
                timestamp_key, timestamp_date, timestamp_interval, write=write
            )
        except (TUFError, SSLibError) as e:
            raise MetadataUpdateError("all", str(e))

    def update_targets_from_keystore(
        self, targets_key, start_date=datetime.datetime.now(), interval=None, write=True
    ):
        """Update targets metadata. Sign it with a key from the file system

        Args:
        - targets_key(securesystemslib.formats.RSAKEY_SCHEMA): Targets key.
        - start_date(datetime): Date to which the specified interval is added when
                                calculating expiration date. If a value is not
                                provided, it is set to the current time
        - interval(int): Number of days added to the start date. If not provided,
                        the default value is used
        - write(bool): If True timestmap metadata will be signed and written

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - TimestampMetadataUpdateError: If any other error happened during metadata update
        """
        try:
            self._try_load_metadata_key("targets", targets_key)
            self._update_metadata("targets", start_date, interval, write=write)
        except (TUFError, SSLibError) as e:
            raise TimestampMetadataUpdateError(str(e))

    def update_targets(
        self,
        targets_key_pin,
        targets_data=None,
        start_date=datetime.datetime.now(),
        interval=None,
        write=True,
        public_key=None,
    ):
        """Update target data, sign with smart card and write.

        Args:
        - targets_key_pin(str): Targets key pin
        - targets_data(dict): (Optional) Dictionary with targets data
        - start_date(datetime): Date to which the specified interval is added when
                                calculating expiration date. If a value is not
                                provided, it is set to the current time
        - interval(int): Number of days added to the start date. If not provided,
                        the default value is used
        - write(bool): If True targets metadata will be signed and written
        - public_key(securesystemslib.formats.RSAKEY_SCHEMA): RSA public key dict

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - MetadataUpdateError: If any other error happened during metadata update
        """
        try:
            if public_key is None:
                from taf.yubikey import get_piv_public_key_tuf

                public_key = get_piv_public_key_tuf()

            if not self.is_valid_metadata_yubikey("targets", public_key):
                raise InvalidKeyError("targets")

            if targets_data:
                self.add_targets(targets_data)

            self.set_metadata_expiration_date("targets", start_date, interval)

            self._repository.targets.add_external_signature_provider(
                public_key,
                partial(
                    targets_signature_provider, public_key["keyid"], targets_key_pin
                ),
            )
            if write:
                self._repository.write("targets")

        except (YubikeyError, TUFError, SSLibError) as e:
            raise TargetsMetadataUpdateError(str(e))

    def update_timestamp(
        self,
        timestamp_key,
        start_date=datetime.datetime.now(),
        interval=None,
        write=True,
    ):
        """Update timestamp metadata.

        Args:
        - timestamp_key
        - start_date(datetime): Date to which the specified interval is added when
                                calculating expiration date. If a value is not
                                provided, it is set to the current time
        - interval(int): Number of days added to the start date. If not provided,
                        the default value is used
        - write(bool): If True timestmap metadata will be signed and written

        Returns:
        None

        Raises:
        - InvalidKeyError: If wrong key is used to sign metadata
        - TimestampMetadataUpdateError: If any other error happened during metadata update
        """
        try:
            self._try_load_metadata_key("timestamp", timestamp_key)
            self._update_metadata("timestamp", start_date, interval, write=write)
        except (TUFError, SSLibError) as e:
            raise TimestampMetadataUpdateError(str(e))

    def writeall(self):
        """Write all dirty metadata files.

        Args:
        None

        Returns:
        None

        Raises:
        - tuf.exceptions.UnsignedMetadataError: If any of the top-level and delegated roles do not
                                                have the minimum threshold of signatures.
        """
        self._repository.writeall()



"""Add platform-dependent libraries to the path.
"""
import os
import sys
from pathlib import Path

_PLATFORM = sys.platform

_PLATFORM_LIBS = str((Path(__file__).parent / "libs").resolve())


def _set_env(env_name, path):
    try:
        os.environ[env_name] += os.pathsep + path
    except KeyError:
        os.environ[env_name] = path


if _PLATFORM == "darwin":
    _set_env("DYLD_LIBRARY_PATH", _PLATFORM_LIBS)
elif _PLATFORM == "linux":
    _set_env("LD_LIBRARY_PATH", _PLATFORM_LIBS)
elif _PLATFORM == "win32":
    _set_env("PATH", _PLATFORM_LIBS)
else:
    raise Exception(f'Platform "{_PLATFORM}" is not supported!')



import json
from pathlib import Path
from subprocess import CalledProcessError

from taf.log import taf_logger
from taf.exceptions import InvalidOrMissingMetadataError, RepositoriesNotFoundError
from taf.git import NamedGitRepository

# {
#     'authentication_repo_name': {
#         'commit' : {
#             'path1': git_repository1
#             'path2': target_git_repository2
#             ...
#         }
#     }
# }

_repositories_dict = {}
repositories_path = "targets/repositories.json"
targets_path = "metadata/targets.json"


def clear_repositories_db():
    global _repositories_dict
    _repositories_dict.clear()


def load_repositories(
    auth_repo,
    repo_classes=None,
    factory=None,
    root_dir=None,
    only_load_targets=False,
    commits=None,
):
    """
    Creates target repositories by reading repositories.json and targets.json files
    at the specified revisions, given an authentication repo.
    If the the commits are not specified, targets will be created based on the HEAD pointer
    of the authentication repository. It is possible to specify git repository class that
    will be created per target.
    Args:
        auth_repo: the authentication repository
        target_classes: a single git repository class, or a dictionary whose keys are
        target paths and values are git repository classes. E.g:
        {
            'path1': GitRepo1,
            'path2': GitRepo2,
            'default': GitRepo3
        }
        When determening a target's class, in case when targets_classes is a dictionary,
        it is first checked if its path is in a key in the dictionary. If it is not found,
        it is checked if default class is set, by looking up value of 'default'. If nothing
        is found, the class is set to TAF's NamedGitRepository.
        If target_classes is a single class, all targets will be of that type.
        If target_classes is None, all targets will be of TAF's NamedGitRepository type.
        root_dir: root directory relative to which the target paths are specified
        commits: Authentication repository's commits at which to read targets.json
        only_load_targets: specifies if only repositories specified in targets.json should be loaded.
        If set to false, all repositories defined in repositories.json are loaded, regardless of if
        they are targets or not.
    """
    global _repositories_dict
    if auth_repo.repo_name not in _repositories_dict:
        _repositories_dict[auth_repo.repo_name] = {}

    if commits is None:
        commits = [auth_repo.head_commit_sha()]

    taf_logger.debug(
        "Loading {}'s target repositories at revisions {}",
        auth_repo.repo_name,
        ", ".join(commits),
    )

    if root_dir is None:
        root_dir = Path(auth_repo.repo_path).parent

    for commit in commits:
        repositories_dict = {}
        # check if already loaded
        if commit in _repositories_dict[auth_repo.repo_name]:
            continue

        _repositories_dict[auth_repo.repo_name][commit] = repositories_dict
        try:
            repositories = _get_json_file(auth_repo, repositories_path, commit)
            targets = _get_json_file(auth_repo, targets_path, commit)
        except InvalidOrMissingMetadataError as e:
            taf_logger.warning("Skipping commit {} due to error {}", commit, e)
            continue

        # target repositories are defined in both mirrors.json and targets.json
        repositories = repositories["repositories"]
        targets = targets["signed"]["targets"]
        for path, repo_data in repositories.items():
            urls = repo_data["urls"]
            target = targets.get(path)
            if target is None and only_load_targets:
                continue
            additional_info = _get_custom_data(repo_data, targets.get(path))

            if factory is not None:
                git_repo = factory(root_dir, path, urls, additional_info)
            else:
                git_repo_class = _determine_repo_class(repo_classes, path)
                git_repo = git_repo_class(root_dir, path, urls, additional_info)

            if not isinstance(git_repo, NamedGitRepository):
                raise Exception(
                    f"{type(git_repo)} is not a subclass of NamedGitRepository"
                )

            repositories_dict[path] = git_repo

        taf_logger.debug(
            "Loaded the following repositories at revision {}: {}",
            commit,
            ", ".join(repositories_dict.keys()),
        )


def _determine_repo_class(repo_classes, path):
    # if no class is specified, return the default one
    if repo_classes is None:
        return NamedGitRepository

    # if only one value is specified, that means that all target repositories
    # should be of the same class
    if not isinstance(repo_classes, dict):
        return repo_classes

    if path in repo_classes:
        return repo_classes[path]

    if "default" in repo_classes:
        return repo_classes["default"]

    return NamedGitRepository


def _get_custom_data(repo, target):
    custom = repo.get("custom", {})
    target_custom = target.get("custom") if target is not None else None
    if target_custom is not None:
        custom.update(target_custom)
    return custom


def _get_json_file(auth_repo, path, commit):
    try:
        return auth_repo.get_json(commit, path)
    except CalledProcessError:
        raise InvalidOrMissingMetadataError("{path} not available at revision {commit}")
    except json.decoder.JSONDecodeError:
        raise InvalidOrMissingMetadataError(
            "{path} not a valid json at revision {commit}"
        )


def get_repositories_paths_by_custom_data(auth_repo, commit=None, **custom):
    if not commit:
        commit = auth_repo.head_commit_sha()
    taf_logger.debug(
        "Auth repo {}: finding paths of repositories by custom data {}",
        auth_repo.repo_name,
        custom,
    )
    targets = auth_repo.get_json(commit, targets_path)
    repositories = auth_repo.get_json(commit, repositories_path)
    repositories = repositories["repositories"]
    targets = targets["signed"]["targets"]

    def _compare(path):
        # Check if `custom` dict is subset of targets[path]['custom'] dict
        try:
            return (
                custom.items()
                <= _get_custom_data(repositories[path], targets.get(path)).items()
            )
        except (AttributeError, KeyError):
            return False

    paths = list(filter(_compare, repositories)) if custom else list(repositories)
    if len(paths):
        taf_logger.debug(
            "Auth repo {}: found the following paths {}", auth_repo.repo_name, paths
        )
        return paths
    taf_logger.error(
        "Auth repo {}: repositories associated with custom data {} not found",
        auth_repo.repo_name,
        custom,
    )
    raise RepositoriesNotFoundError(
        f"Repositories associated with custom data {custom} not found"
    )


def get_deduplicated_repositories(auth_repo, commits):
    global _repositories_dict
    taf_logger.debug(
        "Auth repo {}: getting a deduplicated list of repositories", auth_repo.repo_name
    )
    all_repositories = _repositories_dict.get(auth_repo.repo_name)
    if all_repositories is None:
        taf_logger.error(
            "Repositories defined in authentication repository {} have not been loaded",
            auth_repo.repo_name,
        )
        raise RepositoriesNotFoundError(
            "Repositories defined in authentication repository"
            f" {auth_repo.repo_name} have not been loaded"
        )
    repositories = {}
    # persuming that the newest commit is the last one
    for commit in commits:
        if commit not in all_repositories:
            taf_logger.error(
                "Repositories defined in authentication repository {} at revision {} have "
                "not been loaded",
                auth_repo.repo_name,
                commit,
            )
            raise RepositoriesNotFoundError(
                "Repositories defined in authentication repository "
                "{auth_repo.repo_name} at revision {commit} have not been loaded"
            )
        for path, repo in all_repositories[commit].items():
            # will overwrite older repo with newer
            repositories[path] = repo

    taf_logger.debug(
        "Auth repo {}: deduplicated list of repositories {}",
        auth_repo.repo_name,
        ", ".join(repositories.keys()),
    )
    return repositories


def get_repository(auth_repo, path, commit=None):
    return get_repositories(auth_repo, commit)[path]


def get_repositories(auth_repo, commit=None):
    global _repositories_dict
    if commit is None:
        commit = auth_repo.head_commit_sha()
    taf_logger.debug(
        "Auth repo {}: finding repositories defined at commit {}",
        auth_repo.repo_name,
        commit,
    )
    all_repositories = _repositories_dict.get(auth_repo.repo_name)
    if all_repositories is None:
        taf_logger.error(
            "Repositories defined in authentication repository {} have not been loaded",
            auth_repo.repo_name,
        )
        raise RepositoriesNotFoundError(
            "Repositories defined in authentication repository"
            f" {auth_repo.repo_name} have not been loaded"
        )

    repositories = all_repositories.get(commit)
    if repositories is None:
        taf_logger.error(
            "Repositories defined in authentication repository {} at revision {} have "
            "not been loaded",
            auth_repo.repo_name,
            commit,
        )
        raise RepositoriesNotFoundError(
            "Repositories defined in authentication repository "
            f"{auth_repo.repo_name} at revision {commit} have not been loaded"
        )
    taf_logger.debug(
        "Auth repo {}: found the following repositories at revision {}: {}",
        auth_repo.repo_name,
        commit,
        ", ".join(repositories.keys()),
    )
    return repositories


def get_repositories_by_custom_data(auth_repo, commit=None, **custom_data):
    taf_logger.debug(
        "Auth repo {}: finding repositories by custom data {}",
        auth_repo.repo_name,
        custom_data,
    )
    repositories = get_repositories(auth_repo, commit).values()

    def _compare(repo):
        # Check if `custom` dict is subset of targets[path]['custom'] dict
        try:
            return custom_data.items() <= repo.additional_info.items()
        except (AttributeError, KeyError):
            return False

    found_repos = (
        list(filter(_compare, repositories)) if custom_data else list(repositories)
    )

    if len(found_repos):
        taf_logger.debug(
            "Auth repo {}: found the following repositories {}",
            auth_repo.repo_name,
            repositories,
        )
        return found_repos
    taf_logger.error(
        "Auth repo {}: repositories associated with custom data {} not found",
        auth_repo.repo_name,
        custom_data,
    )
    raise RepositoriesNotFoundError(
        f"Repositories associated with custom data {custom_data} not found"
    )
